%{
    #include "du4.h"
    #include "du4lval.hpp"
    #include "duerr.h"
    #include "du4tok.h"
    #include "du4sem.h"


    /* semnatic value */	
    #define YYSTYPE mlc::MlaskalLval

	/* locations */
	#define YYLTYPE unsigned
	#define YYLLOC_DEFAULT(cur, rhs, n)	do { if(n) (cur)=(rhs)[1]; else (cur)=(rhs)[0]; } while(0)

    /* local stuff */
    using namespace mlc;
    
	/* error stuff */
    #define YYERROR_VERBOSE 1
    static void yyerror(YYLTYPE *line, mlc::MlaskalCtx *ctx, const char *m);
    
%}

%token DUTOK_PROGRAM			    /* program */
%token DUTOK_LABEL			    /* label */
%token DUTOK_CONST			    /* const */
%token DUTOK_TYPE			    /* type */
%token DUTOK_VAR			    /* var */
%token DUTOK_BEGIN			    /* begin */
%token DUTOK_END			    /* end */
%token DUTOK_PROCEDURE			    /* procedure */
%token DUTOK_FUNCTION			    /* function */
%token DUTOK_ARRAY			    /* array */
%token DUTOK_OF				    /* of */
%token DUTOK_GOTO			    /* goto */
%token DUTOK_IF				    /* if */
%token DUTOK_THEN			    /* then */
%token DUTOK_ELSE			    /* else */
%token DUTOK_WHILE			    /* while */
%token DUTOK_DO				    /* do */
%token DUTOK_REPEAT			    /* repeat */
%token DUTOK_UNTIL			    /* until */
%token DUTOK_FOR			    /* for */
%token DUTOK_OR				    /* or */
%token DUTOK_NOT			    /* not */
%token DUTOK_RECORD			    /* record */

/* literals */
%token<id_ci_> DUTOK_IDENTIFIER			    /* identifier */
%token<int_ci_> DUTOK_UINT			    /* unsigned integer */
%token<real_ci_> DUTOK_REAL			    /* real number */
%token<str_ci_> DUTOK_STRING			    /* string */

/* delimiters */
%token DUTOK_SEMICOLON			    /* ; */
%token DUTOK_DOT			    /* . */
%token DUTOK_COMMA			    /* , */
%token DUTOK_EQ				    /* = */
%token DUTOK_COLON			    /* : */
%token DUTOK_LPAR			    /* ( */
%token DUTOK_RPAR			    /* ) */
%token DUTOK_DOTDOT			    /* .. */
%token DUTOK_LSBRA			    /* [ */
%token DUTOK_RSBRA			    /* ] */
%token DUTOK_ASSIGN			    /* := */

/* grouped operators and keywords */
%token DUTOK_OPER_REL			    /* <, <=, <>, >=, > */
%token<dtge_> DUTOK_OPER_SIGNADD		    /* +, - */
%token DUTOK_OPER_MUL			    /* *, /, div, mod, and */
%token DUTOK_FOR_DIRECTION		    /* to, downto */

%type<tp_> type_val ord_type_noid ord_type_id struc_type_val
%type<tp_vec_> ord_type_seq
%type<dtge_> sign_add_opt
%type<int_ci_> ord_const_id ord_const_noid
%type<id_ci_vec_> identifier_seq
%type<pl_p_> formal_params

%start mlaskal
%error-verbose
%pure-parser
%parse-param {mlc::MlaskalCtx *ctx}
%lex-param {mlc::MlaskalCtx *ctx}
%locations

%%

mlaskal:	  DUTOK_PROGRAM DUTOK_IDENTIFIER DUTOK_SEMICOLON program_block DUTOK_DOT { } 
		;

program_block:	  label_block const_block type_block var_block method_block DUTOK_BEGIN stmt_seq DUTOK_END
		;

label_block:	  DUTOK_LABEL uint_seq DUTOK_SEMICOLON  
		| /*no label block */
		;

const_block:	  DUTOK_CONST const_seq
		| /*no const block */
		;

type_block:	  DUTOK_TYPE type_seq
		| /*no type block */
		;

var_block:	  DUTOK_VAR var_seq
		| /*no var block */
		;

method_block:	  procedure_header DUTOK_SEMICOLON block DUTOK_SEMICOLON method_block
		| function_header DUTOK_SEMICOLON block DUTOK_SEMICOLON method_block
		| /*no method block */
		;

block:		  label_block const_block type_block var_block DUTOK_BEGIN stmt_seq DUTOK_END {
				ctx->tab->leave(@7);
			}
		;

procedure_header: DUTOK_PROCEDURE DUTOK_IDENTIFIER DUTOK_LPAR formal_params DUTOK_RPAR	{
				ctx->tab->add_proc(@2, $2, $4); // add_proc deallocates list
				ctx->tab->enter(@2, $2);
				
			}
		| DUTOK_PROCEDURE DUTOK_IDENTIFIER	{
				ctx->tab->add_proc(@2, $2, mlc::create_parameter_list()); // add_proc deallocates list
				ctx->tab->enter(@2, $2);
			}
		;

function_header:  DUTOK_FUNCTION DUTOK_IDENTIFIER DUTOK_LPAR formal_params DUTOK_RPAR DUTOK_COLON DUTOK_IDENTIFIER	{ /* scalar type identifier */
					ctx->tab->add_fnc(@2, $2, get_scalar_type(ctx, $7, @7), $4); // add_fun deallocates list
					ctx->tab->enter(@2, $2);
			}
		| DUTOK_FUNCTION DUTOK_IDENTIFIER DUTOK_COLON DUTOK_IDENTIFIER	{ /* scalar type identifier */
					ctx->tab->add_fnc(@2, $2, get_scalar_type(ctx, $4, @4), mlc::create_parameter_list()); // add_fun deallocates list
					ctx->tab->enter(@2, $2);
			}
		;

formal_params:	  DUTOK_VAR identifier_seq DUTOK_COLON DUTOK_IDENTIFIER	{ /* by ref */
				$$ = mlc::create_parameter_list();
				params_append_byref($$, $2, get_type(ctx, $4, @4));
			}
		| identifier_seq DUTOK_COLON DUTOK_IDENTIFIER	{ /* by val */
				$$ = mlc::create_parameter_list();
				params_append_byval($$, $1, get_type(ctx, $3, @3));
			}
		| formal_params DUTOK_SEMICOLON DUTOK_VAR identifier_seq DUTOK_COLON DUTOK_IDENTIFIER	{ /* by ref */
				$$ = $1;
				params_append_byref($$, $4, get_type(ctx, $6, @6));
			}
		| formal_params DUTOK_SEMICOLON identifier_seq DUTOK_COLON DUTOK_IDENTIFIER { /* by val */
				$$ = $1;
				params_append_byval($$, $3, get_type(ctx, $5, @5));
			}
		;

real_params:	expr
		| expr DUTOK_COMMA real_params
		;

/* various sequences */

uint_seq:	  DUTOK_UINT { ctx->tab->add_label_entry(@1, $1, mlc::new_label( ctx)); } /* unique for labes, thus here */
		| DUTOK_UINT DUTOK_COMMA uint_seq { ctx->tab->add_label_entry(@1, $1, mlc::new_label( ctx)); } /* dtto */
		;

stmt_seq:	  stmt
		| stmt DUTOK_SEMICOLON stmt_seq
		;

matched_seq:	  matched
		| matched DUTOK_SEMICOLON matched_seq
		;

unmatched_seq:	  unmatched
		| unmatched DUTOK_SEMICOLON unmatched_seq
		;

const_seq:	  DUTOK_IDENTIFIER DUTOK_EQ sign_add_opt DUTOK_UINT DUTOK_SEMICOLON {
				mlc::make_const_int(ctx, $1, $3, $4, @1);
			}
		| DUTOK_IDENTIFIER DUTOK_EQ sign_add_opt DUTOK_REAL DUTOK_SEMICOLON {
				mlc::make_const_real(ctx, $1, $3, $4, @1);
			}
		| DUTOK_IDENTIFIER DUTOK_EQ DUTOK_STRING DUTOK_SEMICOLON {
			  	ctx->tab->add_const_str(@1, $1, $3);
		  	}
		| DUTOK_IDENTIFIER DUTOK_EQ DUTOK_IDENTIFIER DUTOK_SEMICOLON /* of konst */ {
				mlc::make_const_from_id(ctx, $1, $3, @1, @3);
			}
		| const_seq DUTOK_IDENTIFIER DUTOK_EQ sign_add_opt DUTOK_UINT DUTOK_SEMICOLON	{
				mlc::make_const_int(ctx, $2, $4, $5, @2);
			}
		| const_seq DUTOK_IDENTIFIER DUTOK_EQ sign_add_opt DUTOK_REAL DUTOK_SEMICOLON {
				mlc::make_const_real(ctx, $2, $4, $5, @2);
			}
		| const_seq DUTOK_IDENTIFIER DUTOK_EQ DUTOK_STRING DUTOK_SEMICOLON	{
			  	ctx->tab->add_const_str(@2, $2, $4);
		  	}
		| const_seq DUTOK_IDENTIFIER DUTOK_EQ DUTOK_IDENTIFIER DUTOK_SEMICOLON	{
				mlc::make_const_from_id(ctx, $2, $4, @2, @4);
			}
		;

type_seq:	  DUTOK_IDENTIFIER DUTOK_EQ type_val DUTOK_SEMICOLON	{
				ctx->tab->add_type(@1, $1, $3);
			}
		| type_seq DUTOK_IDENTIFIER DUTOK_EQ type_val DUTOK_SEMICOLON	{
				ctx->tab->add_type(@2, $2, $4);
			}
		;

var_seq:	  identifier_seq DUTOK_COLON type_val DUTOK_SEMICOLON	{
				var_from_vec(ctx, $1, $3);
			}
		| identifier_seq DUTOK_COLON type_val DUTOK_SEMICOLON var_seq	{
				var_from_vec(ctx, $1, $3);
			}
		;

identifier_seq:	  DUTOK_IDENTIFIER	{
				$$.push_back(std::make_pair< mlc::ls_id_type::const_pointer, int>($1,@1));
			}

		| DUTOK_IDENTIFIER DUTOK_COMMA identifier_seq	{
				$$.push_back(std::make_pair< mlc::ls_id_type::const_pointer, int>($1,@1));
				$$.insert($$.end(), $3.begin(), $3.end());
			}
		;


ordinal_expr_seq: expr /* expr of ordinal type */
		| expr DUTOK_COMMA ordinal_expr_seq
		;

term_seq:	  term
		| term DUTOK_OPER_SIGNADD term_seq
		| term DUTOK_OR term_seq
		;

/* various values */

const_nosign:	  DUTOK_UINT
		| DUTOK_REAL
		| DUTOK_STRING
		;

/* type machinery */
type_val:	  ord_type_noid		{ $$ = $1; }
		| struc_type_val	{ $$ = $1;}
		| DUTOK_IDENTIFIER	{
				mlc::symbol_pointer sp = ctx->tab->find_symbol($1);
				type_pointer tp = sp->access_typed()->type();
				if (sp->kind() != SKIND_TYPE) { error(DUERR_NOTTYPE, @1, *$1); }
				$$ = tp;
			}
		| DUTOK_IDENTIFIER DUTOK_DOTDOT ord_const_id	{
				$$ = range_from_id_ord_const(ctx, $1, $3, @1);
			}
		;

ord_type_id:	  ord_const_noid DUTOK_DOTDOT ord_const_id {
				$$ = range_from_ord_const_ord_const(ctx, $1, $3, @1);
			}
		| DUTOK_IDENTIFIER DUTOK_DOTDOT ord_const_id	{
				$$ = range_from_id_ord_const(ctx, $1, $3, @1);
			}
		| DUTOK_IDENTIFIER	{
				mlc::symbol_pointer sp = ctx->tab->find_symbol($1);
				type_pointer tp = sp->access_typed()->type();
				if (sp->kind() != SKIND_TYPE) { error(DUERR_NOTTYPE, @1, *$1); }
				$$ = tp;
			}
		;

ord_type_noid:	  ord_const_noid DUTOK_DOTDOT ord_const_id	{
				$$ = range_from_ord_const_ord_const(ctx, $1, $3, @1);
			}
		;

struc_type_val:	  DUTOK_ARRAY DUTOK_LSBRA ord_type_seq DUTOK_RSBRA DUTOK_OF type_val	{
				mlc::type_pointer et; /* index type, element type */;
				et = $6;
				/* ord_type_seq size >= 1 */
				for( tp_vec_riter rit = $3.rbegin(); rit != $3.rend(); ++rit) {
					et = ctx->tab->create_array_type(*rit, et);
				}
				$$ = et;
			}
		;

ord_type_seq:	  ord_type_id	{
				$$.clear();
				$$.push_back($1);
			}
		| ord_type_seq DUTOK_COMMA ord_type_id {
				/* pass range sequence */
				/*for(tp_vec_iter it = $1.begin(); it != $1.end(); ++it) {
					$$.push_back(*it);
				}*/
				$$.push_back($3);				
			}
		;

ord_const_noid:	  DUTOK_OPER_SIGNADD DUTOK_IDENTIFIER	{ /* identifier of const integer */
				$$ = ord_const_from_id(ctx, $1, $2, @2);
			}
		| sign_add_opt DUTOK_UINT	{
				$$ = ord_const_from_uint(ctx, $1, $2);
			}
		;

ord_const_id:	  sign_add_opt DUTOK_IDENTIFIER	{ /* identifier of const integer */
				$$ = ord_const_from_id(ctx, $1, $2, @2);
			}
		| sign_add_opt DUTOK_UINT {
				$$ = ord_const_from_uint(ctx, $1, $2);
			}
		;

/* stmt machinery */

stmt:		  matched
		| unmatched 
		;

matched:	  label_inst_opt DUTOK_IF expr DUTOK_THEN matched DUTOK_ELSE matched
		| label_inst_opt variable DUTOK_ASSIGN expr
		| label_inst_opt DUTOK_IDENTIFIER /* identifier of function or variable */ DUTOK_ASSIGN expr
		| label_inst_opt DUTOK_IDENTIFIER /* identifier of procedure */ 
		| label_inst_opt DUTOK_IDENTIFIER /* identifier of procedure */ DUTOK_LPAR real_params DUTOK_RPAR
		| label_inst_opt DUTOK_GOTO DUTOK_UINT
		| label_inst_opt DUTOK_BEGIN stmt_seq DUTOK_END
		| label_inst_opt DUTOK_REPEAT matched_seq DUTOK_UNTIL expr /* expr of bool */ 
		| label_inst_opt DUTOK_FOR DUTOK_IDENTIFIER /* identifier of ordinal type variable */ DUTOK_ASSIGN expr /* ordinal type expr */ DUTOK_FOR_DIRECTION expr /* expr of ordinal */ DUTOK_DO matched
		| label_inst_opt DUTOK_WHILE expr /* expr of bool */ DUTOK_DO matched 
		| 
		;
		
unmatched:	  label_inst_opt DUTOK_IF expr /* expr of bool */ DUTOK_THEN stmt
		| label_inst_opt DUTOK_IF expr /* expr of bool */ DUTOK_THEN matched DUTOK_ELSE unmatched
		| label_inst_opt DUTOK_FOR DUTOK_IDENTIFIER /* identifier of ordinal type variable */ DUTOK_ASSIGN expr /* ordinal type expr */ DUTOK_FOR_DIRECTION expr /* expr of ordinal */ DUTOK_DO unmatched /* ?? matched ?? */
		| label_inst_opt DUTOK_WHILE expr /* expr of bool */ DUTOK_DO unmatched 
		| label_inst_opt DUTOK_REPEAT unmatched_seq DUTOK_UNTIL expr /* expr of bool */ 
		;

expr:		  simple_expr
		| simple_expr DUTOK_OPER_REL simple_expr
		| simple_expr DUTOK_EQ simple_expr
		;

simple_expr:	  term_seq
		| DUTOK_OPER_SIGNADD term_seq
		;

sign_add_opt:	  DUTOK_OPER_SIGNADD	{ $$ = $1;}
		| /* no sign add */	{ $$ = DUTOKGE_PLUS;}
		;

term:		  faktor
		| term DUTOK_OPER_MUL faktor
		;

faktor:		  const_nosign
		| variable
		| DUTOK_IDENTIFIER /* identifier of variable or const or ... */
		| DUTOK_IDENTIFIER /*identifier of function */ DUTOK_LPAR real_params DUTOK_RPAR
		| DUTOK_LPAR expr DUTOK_RPAR
		| DUTOK_NOT faktor
		;

variable:	  DUTOK_IDENTIFIER /* identifier of array variable */ DUTOK_LSBRA ordinal_expr_seq DUTOK_RSBRA
		| variable /* identifier of array variable */ DUTOK_LSBRA ordinal_expr_seq DUTOK_RSBRA
		;

label_inst_opt:	  DUTOK_UINT DUTOK_COLON 
		| /* no label */
		;

%%

static void yyerror(unsigned *line, MlaskalCtx *, const char *m)
{
    error(DUERR_SYNTAX, *line, m);
}
