%{

	#include "duerr.h"
	#include "du4.h"
	#include "du4tok.h"
	#include "du4g.hpp"
	#include "du4lval.hpp"

	#include <string>
	#include <locale>


	/* portability stuff */
	#if defined _MSC_VER	    /* M$ VS2003 does not recognize long long, VS2005 has long long but no strtoull */
	    typedef unsigned __int64	UINT64;
	    #define C64(n)				n##i64
	    #define STRTOULL			_strtoui64
	#else						/* the rest of the world */
	    typedef unsigned long long	UINT64;
	    #define C64(n)				n##LL
	    #define STRTOULL			strtoull
	#endif

	#define STRTOUPPER(str) { \
		std::locale loc; \
		for(size_t i=0; i<str.length(); ++i) { \
			str[i] = toupper(str[i], loc); \
		} \
	}

	#define INT_RANGE 0x80000000

	std::string strbuf;
	unsigned int comment_level = 0;
	unsigned int line_num = 1;

	//UINT64
	int
	my_int_conv(std::string foo) {
		//greater plausible value has 10 digits
		if(foo.size() > 20) {
			mlc::error(mlc::DUERR_INTOUTRANGE, line_num, yytext);
			foo.erase(0, foo.size() - 20);
		}

		UINT64 bar = STRTOULL(foo.c_str(), NULL, 10);
		UINT64 foobar = bar % INT_RANGE;
		if(foobar != bar) {
			mlc::error(mlc::DUERR_INTOUTRANGE, line_num, yytext);
		}
		return (int) foobar; 
		
	}

%}

%option noyywrap
%option nostdinit
%option never-interactive
%option 8bit
%option nounput

DIGIT			[0-9]
UINT			{DIGIT}+
SIGN			[+-]
E			[eE]
WHITESPACE		[ \t\f\r]
LF			[\n]
LETTER			[a-zA-Z]
UINT_MALSUF		[a-zA-Z'][a-zA-Z']*

%x str comment

%%

	/* string */

'	strbuf.clear(); BEGIN(str);

<str>{
''	strbuf += '\'';

'	{ /* closing ' */
		BEGIN(INITIAL);
		//printf("Correct string: %s\n", strbuf.c_str());

		*l = line_num;
		lv->str_ci_ = ctx->tab->ls_str().add(strbuf.c_str());
		return DUTOK_STRING;
	}

<<EOF>> {
		mlc::error(mlc::DUERR_EOFINSTRCHR, line_num);
		BEGIN(INITIAL);
	}

\r	/* ignore */

\n	{ /* error - unterminated string */

		mlc::error(mlc::DUERR_EOLINSTRCHR, line_num);
		
		BEGIN(INITIAL);
		//printf("Prematurely terminated string: %s\n", strbuf.c_str());

		lv->str_ci_ = ctx->tab->ls_str().add(strbuf.c_str());
		*l = line_num;
		++line_num;
		return DUTOK_STRING;
	}

[^'\n\r]+	{
		char *yptr = yytext;

		while ( *yptr )
			strbuf += *yptr++;
	}

	/*\r	ignore */
}

	/* comments */

\{	{
		BEGIN(comment);
		comment_level++;
	}

\}	mlc::error(mlc::DUERR_UNEXPENDCMT, line_num);

<comment>{

\}	{
		comment_level--;
		if(comment_level == 0) BEGIN(INITIAL);
	}

\n	++line_num;

\{	++comment_level;

[^{}\n]	/* eat up anything */

<<EOF>>	{
		BEGIN(INITIAL);
		/* TODO raise error */
		mlc::error(mlc::DUERR_EOFINCMT, line_num);
	}

}
	/* keywords */

[pP][rR][oO][gG][rR][aA][mM]		*l = line_num; return DUTOK_PROGRAM;	/* program */
[lL][aA][bB][eE][lL]			*l = line_num; return DUTOK_LABEL;	/* label */
[cC][oO][nN][sS][tT]			*l = line_num; return DUTOK_CONST;	/* const */
[tT][yY][pP][eE]			*l = line_num; return DUTOK_TYPE;	/* type */
[vV][aA][rR]				*l = line_num; return DUTOK_VAR;	/* var */
[bB][eE][gG][iI][nN]			*l = line_num; return DUTOK_BEGIN;	/* begin */
[eE][nN][dD]				*l = line_num; return DUTOK_END;	/* end */
[pP][rR][oO][cC][eE][dD][uU][rR][eE]	*l = line_num; return DUTOK_PROCEDURE;	/* procedure */
[fF][uU][nN][cC][tT][iI][oO][nN]	*l = line_num; return DUTOK_FUNCTION;	/* function */
[aA][rR][rR][aA][yY]			*l = line_num; return DUTOK_ARRAY;	/* array */
[oO][fF]				*l = line_num; return DUTOK_OF;	/* of */
[gG][oO][tT][oO]			*l = line_num; return DUTOK_GOTO;	/* goto */
[iI][fF]				*l = line_num; return DUTOK_IF;	/* if */
[tT][hH][eE][nN]			*l = line_num; return DUTOK_THEN;	/* then */
[eE][lL][sS][eE]			*l = line_num; return DUTOK_ELSE;	/* else */
[wW][hH][iI][lL][eE]			*l = line_num; return DUTOK_WHILE;	/* while */
[dD][oO]				*l = line_num; return DUTOK_DO;	/* do */
[rR][eE][pP][eE][aA][tT]		*l = line_num; return DUTOK_REPEAT;	/* repeat */
[uU][nN][tT][iI][lL]			*l = line_num; return DUTOK_UNTIL;	/* until */
[fF][oO][rR]				*l = line_num; return DUTOK_FOR;	/* for */
[oO][rR]				*l = line_num; return DUTOK_OR;	/* or */
[nN][oO][tT]				*l = line_num; return DUTOK_NOT;	/* not */
[rR][eE][cC][oO][rR][dD]		*l = line_num; return DUTOK_RECORD;	/* record */

	/* delimiters */

:=	{
		/* := */
		*l = line_num;
		return DUTOK_ASSIGN;
	}

\.\.	{
		/* .. */
		*l = line_num;
		return DUTOK_DOTDOT;
	}

;	{
		/* ; */
		*l = line_num;
		return DUTOK_SEMICOLON;
	}

\.	{
		/* . */
		*l = line_num;
		return DUTOK_DOT;
	}

,	{
		/* , */
		*l = line_num;
		return DUTOK_COMMA;
	}

=	{
		/* = */
		*l = line_num;
		return DUTOK_EQ;
	}

:	{
		/* : */
		*l = line_num;
		return DUTOK_COLON;
	}

\(	{
		/* ( */
		*l = line_num;
		return DUTOK_LPAR;
	}

\)	{
		/* ) */
		*l = line_num;
		return DUTOK_RPAR;
	}

\[	{
		/* [ */
		*l = line_num;
		return DUTOK_LSBRA;
	}

\]	{
		/* ] */
		*l = line_num;
		return DUTOK_RSBRA;
	}


	/* group tokens */

\<	{
		lv->dtge_ = DUTOKGE_LT;
		*l = line_num;
		return DUTOK_OPER_REL;
	}

>	{
		lv->dtge_ = DUTOKGE_GT;
		*l = line_num;
		return DUTOK_OPER_REL;
	}

\<=	{
		lv->dtge_ = DUTOKGE_LE;
		*l = line_num;
		return DUTOK_OPER_REL;
	}	

>=	{
		lv->dtge_ = DUTOKGE_GE;
		*l = line_num;
		return DUTOK_OPER_REL;
	}

\<>	{
		lv->dtge_ = DUTOKGE_NE;
		*l = line_num;
		return DUTOK_OPER_REL;
	}


-	{
		lv->dtge_ = DUTOKGE_MINUS;
		*l = line_num;
		return DUTOK_OPER_SIGNADD;
	}

\+	{
		lv->dtge_ = DUTOKGE_PLUS;
		*l = line_num;
		return DUTOK_OPER_SIGNADD;
	}

\*	{
		lv->dtge_ = DUTOKGE_ASTERISK;
		*l = line_num;
		return DUTOK_OPER_MUL;
	}

\/	{
		lv->dtge_ = DUTOKGE_SOLIDUS;
		*l = line_num;
		return DUTOK_OPER_MUL;
	}

[dD][iI][vV]	{
			lv->dtge_ = DUTOKGE_DIV;
			*l = line_num;
			return DUTOK_OPER_MUL;
		}

[mM][oO][dD]	{
			lv->dtge_ = DUTOKGE_MOD;
			*l = line_num;
			return DUTOK_OPER_MUL;
		}

[aA][nN][dD]	{
			lv->dtge_ = DUTOKGE_AND;
			*l = line_num;
			return DUTOK_OPER_MUL;
		}

[tT][oO]	{
			lv->dtge_ = DUTOKGE_TO;
			*l = line_num;
			return DUTOK_FOR_DIRECTION;
		}

[dD][oO][wW][nN][tT][oO]	{
			lv->dtge_ = DUTOKGE_DOWNTO;
			*l = line_num;
			return DUTOK_FOR_DIRECTION;
		}

	/* literals */

{UINT} {
		/* unsigned integer */
		lv->int_ci_ = ctx->tab->ls_int().add(my_int_conv(yytext));
		*l = line_num;
		return DUTOK_UINT;
	}

{UINT}((\.{DIGIT}+({E}{SIGN}?{UINT})?)|({E}{SIGN}?{UINT}))	{
		/* real number */
		*l = line_num;
		lv->real_ci_ = ctx->tab->ls_real().add(strtod(yytext, NULL));
		return DUTOK_REAL;
	}

{UINT}[a-zA-Z][a-zA-Z0-9]* { 
		/* malformed unsigned integer */
		mlc::error(mlc::DUERR_BADINT, line_num, yytext);
		lv->int_ci_ = ctx->tab->ls_int().add(my_int_conv(yytext));
		*l = line_num;
		return DUTOK_UINT;
	}

{UINT}((\.{DIGIT}+({E}{SIGN}?{UINT})?)|({E}{SIGN}?{UINT}))[a-zA-Z][a-zA-Z0-9]*	{
		/* malformed real number */
		*l = line_num;
		mlc::error(mlc::DUERR_BADREAL, line_num, yytext);
		lv->real_ci_ = ctx->tab->ls_real().add(strtod(yytext, NULL));
		return DUTOK_REAL;
	}

{UINT}\.({WHITESPACE}|([a-zA-Z][a-zA-Z0-9]*)|^) {
		*l = line_num;
		mlc::error(mlc::DUERR_BADREAL, line_num, yytext);
		lv->real_ci_ = ctx->tab->ls_real().add(strtod(yytext, NULL));
		return DUTOK_REAL;
	}


{LETTER}({LETTER}|{DIGIT})*	{
	/* identifier */
	std::string id(yytext);
	STRTOUPPER(id);

	lv->id_ci_ = ctx->tab->ls_id().add( id);
	*l = line_num;
	return DUTOK_IDENTIFIER;
}


	/* junk */

{WHITESPACE}+		/* go out with whitespaces */

{LF}			++line_num;

.			mlc::error(mlc::DUERR_UNKCHAR, line_num, *yytext, *yytext);

%%
