%{
	/* frantisek haas */

	#include "duerr.h"
	#include "du4.h"
	#include "du4tok.h"
	#include "du4g.hpp"
	#include "du4lval.hpp"

	#include <string>
	#include <vector>
	
	/* portability stuff */
	#if defined _MSC_VER	    /* M$, VS2003 does not recognize long long */
	    typedef unsigned __int64	UINT64;
	    #define C64(n)				n##i64
	    #define STRTOULL			_strtoui64
	#else						/* the rest of the world; hmm, SunCC doesn't have strupr, hopefully no one will use it */
	    typedef unsigned long long	UINT64;
	    #define C64(n)				n##LL
	    #define STRTOULL			strtoull
	#endif
	
	using namespace std;
	using namespace mlc;

	const UINT64 MLASKAL_MAX_UINT = 2147483648;

	std::string str = "";
	unsigned cmt = 0;
	unsigned line = 1;
	
	void mng_id( MlaskalLval *lv, MlaskalCtx *ctx, const std::string & strin) {
		string id = string( strin);

		for ( size_t i = 0; i < strin.length(); ++i)
			id[ i] = (char)toupper( strin[ i]);

		lv->id_ci_ = ctx->tab->ls_id().add( id);
	}

	void mng_str( MlaskalLval *lv, MlaskalCtx *ctx, const std::string & strin) {
		lv->str_ci_ = ctx->tab->ls_str().add( strin);
	}	

	void mng_uint( MlaskalLval *lv, MlaskalCtx *ctx, const std::string & strin) {						
		UINT64 cuint = STRTOULL( strin.c_str(), NULL, 0);
		unsigned rcuint = cuint % MLASKAL_MAX_UINT;		

		if ( cuint != rcuint)
			error( DUERR_INTOUTRANGE, line, strin);

		lv->int_ci_ = ctx->tab->ls_int().add( rcuint);
		
	}		

	void mng_baduint( MlaskalLval *lv, MlaskalCtx *ctx, const std::string & strin) {
		error( DUERR_BADINT, line, strin);

		UINT64 cuint = STRTOULL( strin.c_str(), NULL, 0);
		unsigned rcuint = cuint % MLASKAL_MAX_UINT;		

		// ?
		lv->int_ci_ = ctx->tab->ls_int().add( rcuint);

	}

	void mng_real( MlaskalLval *lv, MlaskalCtx *ctx, const std::string & strin) {
		double creal = strtod( strin.c_str(), NULL);
		lv->real_ci_ = ctx->tab->ls_real().add( creal);
	}

	void mng_badreal( MlaskalLval *lv, MlaskalCtx *ctx, const std::string & strin) {
		error( DUERR_BADREAL, line, strin);
		double creal = strtod( strin.c_str(), NULL);
		// ?
		lv->real_ci_ = ctx->tab->ls_real().add( creal);
	}
		
%}

%option noyywrap
%option nostdinit
%option never-interactive
%option 8bit
%option nounput

/* defines */

PROGRAM			[pP][rR][oO][gG][rR][aA][mM]
LABEL			[lL][aA][bB][eE][lL]
CONST			[cC][oO][nN][sS][tT]
TYPE			[tT][yY][pP][eE]
VAR				[vV][aA][rR]
BEGIN			[bB][eE][gG][iI][nN]
END				[eE][nN][dD]
PROCEDURE		[pP][rR][oO][cC][eE][dD][uU][rR][eE]
FUNCTION		[fF][uU][nN][cC][tT][iI][oO][nN]
ARRAY			[aA][rR][rR][aA][yY]
OF				[oO][fF]
GOTO			[gG][oO][tT][oO]
IF				[iI][fF]
THEN			[tT][hH][eE][nN]
ELSE			[eE][lL][sS][eE]
WHILE			[wW][hH][iI][lL][eE]
DO				[dD][oO]			
REPEAT			[rR][eE][pP][eE][aA][tT]
UNTIL			[uU][nN][tT][iI][lL]
FOR				[fF][oO][rR]
OR				[oO][rR]
NOT				[nN][oO][tT]
RECORD			[rR][eE][cC][oO][rR][dD]

DIV				[dD][iI][vV]
MOD				[mM][oO][dD]
AND				[aA][nN][dD]
TO				[tT][oO]
DOWNTO			[dD][oO][wW][nN][tT][oO]

IDENTIFIER		[A-Za-z][A-Za-z0-9]*
DIGIT			[0-9]
UINT			{DIGIT}+
BADUINT			{UINT}[A-Za-z]+
REALP1			{UINT}\.{DIGIT}+
REALP2			{UINT}\.{DIGIT}+[eE][+-]?{UINT}
REALP3			{UINT}[eE][+-]?{UINT}
REAL			{REALP1}|{REALP2}|{REALP3}
BADREALP1		{UINT}\.{DIGIT}+[eE][+-]
BADREALP2		{UINT}\.{DIGIT}+[eE][A-Za-z][A-Za-z0-9]*
BADREAL			{BADREALP1}|{BADREALP2}
WHITESPACE		[ \r\t\f]

%x				STR
%x				CMT

%%

<STR>{
	<<EOF>>			error( DUERR_EOFINSTRCHR, line);
	(\n)			BEGIN( INITIAL); error( DUERR_EOLINSTRCHR, line); mng_str( lv, ctx, str); *l = line; ++line; return DUTOK_STRING;
	(\')			BEGIN( INITIAL); mng_str( lv, ctx, str); *l = line; return DUTOK_STRING;
	(\'\')			str += '\'';
	[^\n\'] 		str += *yytext;	
}

<CMT>{
	<<EOF>>			BEGIN( INITIAL); error( DUERR_EOFINCMT, line);
	(\n)			++line;
	(\{)			++cmt;
	(\})			if ( cmt == 0) {
						BEGIN( INITIAL);
					} else {
						--cmt;
					}
	[^\n\{\}]   	/* trash */
}

{WHITESPACE}+		/* trash */
(\n)				++line;

(\')				BEGIN( STR); str = ""; 
(\{)				BEGIN( CMT); cmt = 0; 
(\})				error( DUERR_UNEXPENDCMT, line);

{PROGRAM}			*l = line; return DUTOK_PROGRAM;
{LABEL}				*l = line; return DUTOK_LABEL;
{CONST}				*l = line; return DUTOK_CONST;
{TYPE}				*l = line; return DUTOK_TYPE;
{VAR}				*l = line; return DUTOK_VAR;
{BEGIN}				*l = line; return DUTOK_BEGIN;
{END}				*l = line; return DUTOK_END;
{PROCEDURE}			*l = line; return DUTOK_PROCEDURE;
{FUNCTION}			*l = line; return DUTOK_FUNCTION;
{ARRAY}				*l = line; return DUTOK_ARRAY;
{OF}				*l = line; return DUTOK_OF;
{GOTO}				*l = line; return DUTOK_GOTO;
{IF}				*l = line; return DUTOK_IF;
{THEN}				*l = line; return DUTOK_THEN;
{ELSE}				*l = line; return DUTOK_ELSE;
{WHILE}				*l = line; return DUTOK_WHILE;
{DO}				*l = line; return DUTOK_DO;
{REPEAT}			*l = line; return DUTOK_REPEAT;
{UNTIL} 			*l = line; return DUTOK_UNTIL;
{FOR}				*l = line; return DUTOK_FOR;
{OR}				*l = line; return DUTOK_OR;
{NOT}				*l = line; return DUTOK_NOT;
{RECORD}			*l = line; return DUTOK_RECORD;

(;)					*l = line; return DUTOK_SEMICOLON;
(\.)				*l = line; return DUTOK_DOT;
(,)					*l = line; return DUTOK_COMMA;
(=)					*l = line; return DUTOK_EQ;
(:)					*l = line; return DUTOK_COLON;
(\()				*l = line; return DUTOK_LPAR;
(\))				*l = line; return DUTOK_RPAR;
(\.\.)				*l = line; return DUTOK_DOTDOT;
(\[)				*l = line; return DUTOK_LSBRA;
(\])				*l = line; return DUTOK_RSBRA;
(:=)				*l = line; return DUTOK_ASSIGN;

{TO}				*l = line; lv->dtge_ = DUTOKGE_TO; return DUTOK_FOR_DIRECTION;
{DOWNTO}			*l = line; lv->dtge_ = DUTOKGE_DOWNTO; return DUTOK_FOR_DIRECTION;

(\*)				*l = line; lv->dtge_ = DUTOKGE_ASTERISK; return DUTOK_OPER_MUL;
(\/)				*l = line; lv->dtge_ = DUTOKGE_SOLIDUS; return DUTOK_OPER_MUL;
{DIV}				*l = line; lv->dtge_ = DUTOKGE_DIV; return DUTOK_OPER_MUL;
{MOD}				*l = line; lv->dtge_ = DUTOKGE_MOD; return DUTOK_OPER_MUL;
{AND}				*l = line; lv->dtge_ = DUTOKGE_AND; return DUTOK_OPER_MUL;

{IDENTIFIER}		*l = line; mng_id( lv, ctx, yytext); return DUTOK_IDENTIFIER;
{REAL}				*l = line; mng_real( lv, ctx, yytext); return DUTOK_REAL;
{BADREAL}			*l = line; mng_badreal( lv, ctx, yytext); return DUTOK_REAL;
{UINT}				*l = line; mng_uint( lv, ctx, yytext); return DUTOK_UINT;
{BADUINT}			*l = line; mng_baduint( lv, ctx, yytext); return DUTOK_UINT;

(<)					*l = line; lv->dtge_ = DUTOKGE_LT; return DUTOK_OPER_REL;
(<=)				*l = line; lv->dtge_ = DUTOKGE_LE; return DUTOK_OPER_REL;
(<>)				*l = line; lv->dtge_ = DUTOKGE_NE; return DUTOK_OPER_REL;
(>=)				*l = line; lv->dtge_ = DUTOKGE_GE; return DUTOK_OPER_REL;
(>)					*l = line; lv->dtge_ = DUTOKGE_GT; return DUTOK_OPER_REL;

(\+)				*l = line; lv->dtge_ = DUTOKGE_PLUS; return DUTOK_OPER_SIGNADD;
(-)					*l = line; lv->dtge_ = DUTOKGE_MINUS; return DUTOK_OPER_SIGNADD;

.					mlc::error( mlc::DUERR_UNKCHAR, 0, *yytext, *yytext);
