%{
	/* frantisek haas */

    #include "du4.h"
    #include "du4lval.hpp"
    #include "duerr.h"
    #include "du4tok.h"
    #include "du4sem.h"
	#include <iostream>

   /* semantic value */
    #define YYSTYPE mlc::MlaskalLval

	/* locations */
	#define YYLTYPE unsigned
	#define YYLLOC_DEFAULT(cur, rhs, n)	do { if(n) (cur)=(rhs)[1]; else (cur)=(rhs)[0]; } while(0)

    /* local stuff */
    using namespace mlc;
    
	/* error stuff */
    #define YYERROR_VERBOSE 1
    static void yyerror(YYLTYPE *line, mlc::MlaskalCtx *ctx, const char *m);
    
%}

%token DUTOK_PROGRAM			/* program */
%token DUTOK_LABEL			    /* label */
%token DUTOK_CONST			    /* const */
%token DUTOK_TYPE			    /* type */
%token DUTOK_VAR			    /* var */
%token DUTOK_BEGIN			    /* begin */
%token DUTOK_END			    /* end */
%token DUTOK_PROCEDURE			/* procedure */
%token DUTOK_FUNCTION			/* function */
%token DUTOK_ARRAY			    /* array */
%token DUTOK_OF				    /* of */
%token DUTOK_GOTO			    /* goto */
%token DUTOK_IF				    /* if */
%token DUTOK_THEN			    /* then */
%token DUTOK_ELSE			    /* else */
%token DUTOK_WHILE			    /* while */
%token DUTOK_DO				    /* do */
%token DUTOK_REPEAT			    /* repeat */
%token DUTOK_UNTIL			    /* until */
%token DUTOK_FOR			    /* for */
%token DUTOK_OR				    /* or */
%token DUTOK_NOT			    /* not */
%token DUTOK_RECORD			    /* record */

/* literals */
%token<id_ci_>		DUTOK_IDENTIFIER		/* identifier */
%token<int_ci_>		DUTOK_UINT			    /* unsigned integer */
%token<real_ci_>	DUTOK_REAL			    /* real number */
%token<str_ci_>		DUTOK_STRING			/* string */

/* delimiters */
%token DUTOK_SEMICOLON			/* ; */
%token DUTOK_DOT			    /* . */
%token DUTOK_COMMA			    /* , */
%token DUTOK_EQ				    /* = */
%token DUTOK_COLON			    /* : */
%token DUTOK_LPAR			    /* ( */
%token DUTOK_RPAR			    /* ) */
%token DUTOK_DOTDOT			    /* .. */
%token DUTOK_LSBRA			    /* [ */
%token DUTOK_RSBRA			    /* ] */
%token DUTOK_ASSIGN			    /* := */

/* grouped operators and keywords */
%token<dtge_>		DUTOK_OPER_REL			    /* <, <=, <>, >=, > */
%token<dtge_>		DUTOK_OPER_SIGNADD		    /* +, - */
%token<dtge_>		DUTOK_OPER_MUL			    /* *, /, div, mod, and */
%token<dtge_>		DUTOK_FOR_DIRECTION		    /* to, downto */

%type<int_ci_>				const_ordinal_id, const_ordinal_noid
%type<type_ptr_>			type, type_ordinal_id, type_ordinal_noid, type_structured
%type<type_ptr_vec_>		ordinal_type_seq
%type<id_ci_vec_>			identifier_seq
%type<formal_params_>		formal_params

%start mlaskal
%error-verbose
%pure-parser
%parse-param {mlc::MlaskalCtx *ctx}
%lex-param {mlc::MlaskalCtx *ctx}
%locations

%%
	
// --------------------------------------------------------------------------------
// -------------------------------------------------------------------------------- 
// mlaskal	

mlaskal:			DUTOK_PROGRAM DUTOK_IDENTIFIER DUTOK_SEMICOLON program_block DUTOK_DOT
	;
	
program_block:		label_block const_block type_block var_block method_block command_block
		;	

// --------------------------------------------------------------------------------
// --------------------------------------------------------------------------------
// label

label_seq:			DUTOK_UINT
					{
						ctx->tab->add_label_entry( @1, $1, mlc::new_label( ctx));
					}
		|			DUTOK_UINT DUTOK_COMMA label_seq
					{
						ctx->tab->add_label_entry( @1, $1, mlc::new_label( ctx));
					}
		;

label_block:		DUTOK_LABEL label_seq DUTOK_SEMICOLON
		|			/* lambda */
		;		
		
// -------------------------------------------------------------------------------- 
// -------------------------------------------------------------------------------- 
// const
	
const_seq:			DUTOK_IDENTIFIER DUTOK_EQ DUTOK_UINT DUTOK_SEMICOLON
					{
						ctx->tab->add_const_int( @1, $1, $3);
					}
		|			const_seq DUTOK_IDENTIFIER DUTOK_EQ DUTOK_UINT DUTOK_SEMICOLON 
					{
						ctx->tab->add_const_int( @2, $2, $4);
					}
		|			DUTOK_IDENTIFIER DUTOK_EQ DUTOK_REAL DUTOK_SEMICOLON
					{
						ctx->tab->add_const_real( @1, $1, $3);
					}
		|			const_seq DUTOK_IDENTIFIER DUTOK_EQ DUTOK_REAL DUTOK_SEMICOLON 
					{
						ctx->tab->add_const_real( @2, $2, $4);
					}
		|			DUTOK_IDENTIFIER DUTOK_EQ DUTOK_STRING DUTOK_SEMICOLON
					{
						ctx->tab->add_const_str( @1, $1, $3);
					}
		|			const_seq DUTOK_IDENTIFIER DUTOK_EQ DUTOK_STRING DUTOK_SEMICOLON 
					{
						ctx->tab->add_const_str( @2, $2, $4);
					}
		|			DUTOK_IDENTIFIER DUTOK_EQ DUTOK_OPER_SIGNADD DUTOK_UINT DUTOK_SEMICOLON
					{		
						mlc::add_const_int( ctx, @1, $1, $4, $3);		
					}
		|			const_seq DUTOK_IDENTIFIER DUTOK_EQ DUTOK_OPER_SIGNADD DUTOK_UINT DUTOK_SEMICOLON 
					{
						mlc::add_const_int( ctx, @2, $2, $5, $4);	
					}
		|			DUTOK_IDENTIFIER DUTOK_EQ DUTOK_OPER_SIGNADD DUTOK_REAL DUTOK_SEMICOLON
					{
						mlc::add_const_real( ctx, @1, $1, $4, $3);
					}
		|			const_seq DUTOK_IDENTIFIER DUTOK_EQ DUTOK_OPER_SIGNADD DUTOK_REAL DUTOK_SEMICOLON 
					{
						mlc::add_const_real( ctx, @2, $2, $5, $4);	
					}
		|			DUTOK_IDENTIFIER DUTOK_EQ DUTOK_IDENTIFIER DUTOK_SEMICOLON
					{
						mlc::add_const_id( ctx, @1, $1, $3);
					}
		|			const_seq DUTOK_IDENTIFIER DUTOK_EQ DUTOK_IDENTIFIER DUTOK_SEMICOLON 
					{
						mlc::add_const_id( ctx, @2, $2, $4);
					}
		;
		
const_block:		DUTOK_CONST const_seq
		|			/* lambda */
		;

const_ordinal_id:	DUTOK_UINT		
					{
						$$ = $1;
					}
		|			DUTOK_OPER_SIGNADD DUTOK_UINT	
					{
						$$ =  mlc::sgnlsint_to_lsint( ctx, @1, $2, $1);
					}
		|			DUTOK_IDENTIFIER /* celociselne konstanty */
					{				
						$$ = mlc::lsid_to_lsint( ctx, @1, $1, DUTOKGE_PLUS);
					}
		|			DUTOK_OPER_SIGNADD DUTOK_IDENTIFIER /* celociselne konstanty */
					{
						$$ = mlc::lsid_to_lsint( ctx, @1, $2, $1);
					}	
		;
	
const_ordinal_noid:	DUTOK_UINT		
					{
						$$ = $1;
					}
		|			DUTOK_OPER_SIGNADD DUTOK_UINT	
					{
						$$ = sgnlsint_to_lsint( ctx, @1, $2, $1);
					}
		|			DUTOK_OPER_SIGNADD DUTOK_IDENTIFIER /* celociselne konstanty */		
					{
						$$ = mlc::lsid_to_lsint( ctx, @1, $2, $1);
					}	
		;		
		
// --------------------------------------------------------------------------------
// -------------------------------------------------------------------------------- 
// type

type_ordinal_id:	DUTOK_IDENTIFIER /* ordinalniho typu */
					{
						$$ = mlc::lsid_to_range( ctx, @1, $1);						
					}
		|			const_ordinal_id DUTOK_DOTDOT const_ordinal_id
					{
						$$ = ctx->tab->create_range_type( $1, $3);								
					}
		;

type_ordinal_noid:	DUTOK_IDENTIFIER /* celociselne konstanty */ DUTOK_DOTDOT const_ordinal_id
					{
						$$ = ctx->tab->create_range_type( mlc::lsid_to_lsint( ctx, @1, $1, DUTOKGE_PLUS), $3);
					}
		|			const_ordinal_noid DUTOK_DOTDOT const_ordinal_id
					{
						$$ = ctx->tab->create_range_type( $1, $3);						
					}	
		;
		
ordinal_type_seq:	type_ordinal_id
					{
						$$.push_back( $1);
					}
				|	type_ordinal_id DUTOK_COMMA ordinal_type_seq
					{
						$$.push_back( $1);
						for ( size_t i = 0; i < $3.size(); ++i)
							$$.push_back( $3[ i]);
					}
				;

type_structured:	DUTOK_ARRAY DUTOK_LSBRA ordinal_type_seq DUTOK_RSBRA DUTOK_OF type
					{
						$$ = mlc::create_array( ctx, $3, $6);
					}
		;

type:				DUTOK_IDENTIFIER /* ordinalniho typu - strukturovaneho typu - typu  */
					{
						$$ = mlc::lsid_to_type( ctx, @1, $1);													
					}
	|				type_ordinal_noid
					{
						$$ = $1;
					}
	|				type_structured
					{
						$$ = $1;
					}
	;

type_seq:			DUTOK_IDENTIFIER DUTOK_EQ type DUTOK_SEMICOLON
					{
						ctx->tab->add_type( @1, $1, $3);		
					}
		|			type_seq DUTOK_IDENTIFIER DUTOK_EQ type DUTOK_SEMICOLON 
					{
						ctx->tab->add_type( @2, $2, $4);
					}
		;

type_block:			DUTOK_TYPE type_seq
		|			/* lambda */
		;
		
// --------------------------------------------------------------------------------
// -------------------------------------------------------------------------------- 
// method

block:				label_block const_block type_block var_block command_block
					{
						
					}
	;
	
formal_params:		identifier_seq DUTOK_COLON DUTOK_IDENTIFIER /* typu */ 
					{
						$$ = mlc::create_parameter_list();
						for ( size_t i = 0; i < $1.size(); ++i)
							$$->append_parameter_by_value( $1[ i ], mlc::lsid_to_type( ctx, @3, $3));
					}
		|			DUTOK_VAR identifier_seq DUTOK_COLON DUTOK_IDENTIFIER /* typu */
					{
						$$ = mlc::create_parameter_list();
						for ( size_t i = 0; i < $2.size(); ++i)
							$$->append_parameter_by_reference( $2[ i ], mlc::lsid_to_type( ctx, @4, $4));
					}
		|			identifier_seq DUTOK_COLON DUTOK_IDENTIFIER /* typu */ DUTOK_SEMICOLON formal_params
					{
						$$ = mlc::create_parameter_list();
						for ( size_t i = 0; i < $1.size(); ++i)
							$$->append_parameter_by_value( $1[ i ], mlc::lsid_to_type( ctx, @3, $3));
						$$->append_and_kill( $5);
					}
		|			DUTOK_VAR identifier_seq DUTOK_COLON DUTOK_IDENTIFIER /* typu */ DUTOK_SEMICOLON formal_params
					{
						$$ = mlc::create_parameter_list();
						for ( size_t i = 0; i < $2.size(); ++i)
							$$->append_parameter_by_reference( $2[ i ], mlc::lsid_to_type( ctx, @4, $4));
						$$->append_and_kill( $6);
					}
		;				
	
procedure:			DUTOK_PROCEDURE DUTOK_IDENTIFIER 
					{
						ctx->tab->add_proc( @1, $2, mlc::create_parameter_list());
						ctx->tab->enter( @1, $2);
					}
		|			DUTOK_PROCEDURE DUTOK_IDENTIFIER DUTOK_LPAR formal_params DUTOK_RPAR
					{
						ctx->tab->add_proc( @1, $2, $4);
						ctx->tab->enter( @1, $2);
					}
		;
		
function:			DUTOK_FUNCTION DUTOK_IDENTIFIER DUTOK_COLON type /* scalar */
					{
						ctx->tab->add_fnc( @1, $2, $4, mlc::create_parameter_list());
						ctx->tab->enter( @1, $2);
					}
		|			DUTOK_FUNCTION DUTOK_IDENTIFIER DUTOK_LPAR formal_params DUTOK_RPAR DUTOK_COLON type /* scalar */
					{
						ctx->tab->add_fnc( @1, $2, $7, $4);
						ctx->tab->enter( @1, $2);	
					}
		;
		
method_block:		method_block procedure DUTOK_SEMICOLON block DUTOK_SEMICOLON
					{
						ctx->tab->leave( @2);
					}
		|			method_block function DUTOK_SEMICOLON block DUTOK_SEMICOLON 
					{
						ctx->tab->leave( @2);
					}
		|			/* lambda */
		;
			
// --------------------------------------------------------------------------------
// -------------------------------------------------------------------------------- 
// command

command:			command_m
	|				command_u
	;

command_m:			DUTOK_UINT /* label */ DUTOK_COLON  command_m
	|				variable /* navic id funkce */ DUTOK_ASSIGN expr
	|				DUTOK_IDENTIFIER /* procedury */
	|				DUTOK_IDENTIFIER /* procedury */ real_params_seq 
	|				DUTOK_GOTO DUTOK_UINT /* label */
	|				DUTOK_BEGIN command_seq DUTOK_END
	|				DUTOK_IF expr /* typu boolean */ DUTOK_THEN command_m DUTOK_ELSE command_m
	|				DUTOK_WHILE expr /* typu boolean */ DUTOK_DO command_m
	|				DUTOK_REPEAT command_seq DUTOK_UNTIL expr /* typu boolean */
	|				DUTOK_FOR DUTOK_IDENTIFIER /* ordinalniho typu */ DUTOK_ASSIGN expr /* ordinalniho typu */ DUTOK_FOR_DIRECTION expr /* ordinalniho typu */ DUTOK_DO command_m
	|				/* lambda nebo ; ? */
	;	
	
command_u:			DUTOK_IF expr /* typu boolean */ DUTOK_THEN command 
	|				DUTOK_IF expr /* typu boolean */ DUTOK_THEN command_m DUTOK_ELSE command_u
	;	
		
command_seq:		command
		|			command DUTOK_SEMICOLON command_seq
		;
		
command_block:		DUTOK_BEGIN command_seq DUTOK_END
		;
		
// --------------------------------------------------------------------------------
// -------------------------------------------------------------------------------- 
// var

identifier_seq:		DUTOK_IDENTIFIER
					{
						$$.push_back( $1);
					}
			|		DUTOK_IDENTIFIER DUTOK_COMMA identifier_seq
					{
						$$.push_back( $1);
						for ( size_t i = 0; i < $3.size(); ++i)
							$$.push_back( $3[ i ]);
					}
			;

var_seq:			identifier_seq DUTOK_COLON type DUTOK_SEMICOLON
					{
						for ( size_t i = 0; i < $1.size(); ++i)
							ctx->tab->add_var( @2, $1[ i ], $3);
					}
	|				identifier_seq DUTOK_COLON type DUTOK_SEMICOLON var_seq
					{
						for ( size_t i = 0; i < $1.size(); ++i)
							ctx->tab->add_var( @2, $1[ i ], $3);
					}
	;

var_block:			DUTOK_VAR var_seq
		|			/* lambda */
		;	
		
// --------------------------------------------------------------------------------
// -------------------------------------------------------------------------------- 
// variable

variable:			DUTOK_IDENTIFIER /* promenne */
		|			variable /* promenna typu pole */ DUTOK_LSBRA expr_seq DUTOK_RSBRA
		;
		
// --------------------------------------------------------------------------------
// --------------------------------------------------------------------------------
// expr, simple expr, term, faktor

real_params:		expr	
		|			expr DUTOK_COMMA real_params
		;
	
real_params_seq:	DUTOK_LPAR real_params DUTOK_RPAR
			;

faktor_const_nosgn:	DUTOK_UINT
				|	DUTOK_REAL
				|	DUTOK_STRING
				; 			

faktor:				faktor_const_nosgn
	|				variable /* navic identifikator promenne */
	|				DUTOK_IDENTIFIER /* funkce */ real_params_seq
	|				DUTOK_LPAR expr DUTOK_RPAR
	|				DUTOK_NOT faktor
	;

term:				faktor
	|				faktor DUTOK_OPER_MUL term
	;	
	
term_seq:			term
		|			term DUTOK_OPER_SIGNADD term_seq
		;

expr_simple:		term_seq
		|			DUTOK_OPER_SIGNADD term_seq
		|			DUTOK_OR term_seq
		;

expr:				expr_simple
	|				expr_simple DUTOK_OPER_REL expr_simple
	|				expr_simple DUTOK_EQ expr_simple
	;
	
expr_seq:			expr /* ordinalniho typu */ 
		|			expr /* ordinalniho typu */ DUTOK_COMMA expr_seq
		;

// --------------------------------------------------------------------------------
// --------------------------------------------------------------------------------
// --------------------------------------------------------------------------------

%%

static void yyerror(unsigned *line, MlaskalCtx *, const char *m)
{
    error(DUERR_SYNTAX, *line, m);
}
