/* frantisek haas */

/*

DU4SEM.CPP

JY

Mlaskal's semantic interface for DU4

*/

#include "du4sem.h"
#include "du4tok.h"
#include "du4.h"
#include "duerr.h"

namespace mlc {

	void add_const_int( MlaskalCtx *ctx, int line, mlc::ls_id_index idx, mlc::ls_int_index symidx, int dtge)
	{
		int val = * symidx;

		if ( dtge == DUTOKGE_PLUS) {
			ctx->tab->add_const_int( line, idx, symidx);
		} else if ( dtge == DUTOKGE_MINUS) {
			val = -val;
			ctx->tab->add_const_int( line, idx, ctx->tab->ls_int().add( val));
		} else 	{

		}	
	}

	void add_const_real( MlaskalCtx *ctx, int line, mlc::ls_id_index idx, mlc::ls_real_index symidx, int dtge)
	{
		double val = * symidx;

		if ( dtge == DUTOKGE_PLUS) {
			ctx->tab->add_const_real( line, idx, symidx);
		} else if ( dtge == DUTOKGE_MINUS) {
			val = -val;
			ctx->tab->add_const_real( line, idx, ctx->tab->ls_real().add( val));
		} else 	{
			assert(false);
		}	
	}	

	void add_const_id( MlaskalCtx *ctx, int line, mlc::ls_id_index idx, mlc::ls_id_index symidx)
	{
		mlc::symbol_pointer sp = ctx->tab->find_symbol( symidx);
				
		if ( sp->kind() != SKIND_CONST ) { 
			error( DUERR_NOTCONST, line, *symidx); 
		}
		
		mlc::ls_int_index ls_int;
		mlc::ls_real_index ls_real;
		mlc::ls_str_index ls_str;
		bool val_bool;

		switch ( sp->access_typed()->type()->cat())
		{
		case mlc::TCAT_INT:
				ls_int = sp->access_const()->access_int_const()->int_value();
				ctx->tab->add_const_int( line, idx, ls_int);
			break;

		case mlc::TCAT_REAL:
				ls_real = sp->access_const()->access_real_const()->real_value();
				ctx->tab->add_const_real( line, idx, ls_real);
			break;

		case mlc::TCAT_STR:
				ls_str = sp->access_const()->access_str_const()->str_value();
				ctx->tab->add_const_str( line, idx, ls_str);
			break;

		case mlc::TCAT_BOOL:
			    val_bool = sp->access_const()->access_bool_const()->bool_value();
				ctx->tab->add_const_bool( line, idx, val_bool);
			break;

		default:
			assert(false);
			break;
		}
	}

	mlc::type_pointer create_array( MlaskalCtx *ctx, std::vector< mlc::type_pointer> type_ptr_vec, mlc::type_pointer type_ptr)
	{		
		mlc::type_pointer tp = ctx->tab->create_array_type( type_ptr_vec[ type_ptr_vec.size() - 1 ], type_ptr);
		mlc::type_pointer ltp = tp;

		int i = type_ptr_vec.size() - 2;
						
		while ( i >= 0)	{
			tp = ctx->tab->create_array_type( type_ptr_vec[ i ], ltp);
			ltp = tp;
			--i;
		}

		return ltp;
	}

	mlc::ls_int_index sgnlsint_to_lsint( MlaskalCtx *ctx, int line, mlc::ls_int_index sgnlsint, int dtge)
	{
		mlc::ls_int_index lsint;
		
		if ( dtge == DUTOKGE_PLUS)
			lsint = sgnlsint;
		else {
			int val = *sgnlsint;
			lsint = ctx->tab->ls_int().add( -val);
		}

		return lsint;
	}
	
	mlc::ls_int_index lsid_to_lsint( MlaskalCtx *ctx, int line, mlc::ls_id_index lsid, int dtge)
	{
		mlc::ls_int_index lsint;
		mlc::symbol_pointer sp = ctx->tab->find_symbol( lsid);			
											
		if ( !sp || (sp->kind() != SKIND_CONST)) { 
			error( DUERR_NOTCONST, line, *lsid); 

		} else { 

			if ( dtge == DUTOKGE_PLUS) {
				lsint = sp->access_const()->access_int_const()->int_value(); 

			} else if ( dtge == DUTOKGE_MINUS) {
				mlc::ls_int_index tmp = sp->access_const()->access_int_const()->int_value();
				int val = *tmp;
				val = -val;
				lsint = ctx->tab->ls_int().add( val);

			} else 	{
				assert(false);
			}	
			
		}	

		return lsint;
	}

	mlc::type_pointer lsid_to_range( MlaskalCtx *ctx, int line, mlc::ls_id_index lsid)
	{
		mlc::symbol_pointer sp = ctx->tab->find_symbol( lsid);
		mlc::type_pointer tp;

		if ( !sp || (sp->kind() != SKIND_TYPE)) { 
			error( DUERR_NOTTYPE, line, *lsid); 
		} else { 
			tp = sp->access_type()->type(); 
		}			

		return tp;
	}

	mlc::type_pointer lsid_to_type( MlaskalCtx *ctx, int line, mlc::ls_id_index lsid)
	{
		mlc::symbol_pointer sp = ctx->tab->find_symbol( lsid);
		mlc::type_pointer tp;

		if ( !sp || (sp->kind() != SKIND_TYPE)) { 
			error( DUERR_NOTTYPE, line, *lsid); 
		} else { 
			tp = sp->access_type()->type(); 
		}			

		return tp;
	}

}
