/*
 *	frantisek farka
 */

#include "du5sem.h"
#include "du5tok.h"
#include "du5.h"
#include "duerr.h"

namespace mlc {


	void 
	make_const_from_id(MlaskalCtx* ctx, mlc::ls_id_type::const_pointer id, mlc::ls_id_type::const_pointer val, int l_id, int l_val) {

		mlc::symbol_pointer sp = ctx->tab->find_symbol(val);
		if ( sp->kind() != SKIND_CONST ) {
			error( DUERR_NOTCONST, l_val, *val);
		} else {
			bool bval;
			mlc::ls_int_index iptr;
			mlc::ls_real_index rptr;
			mlc::ls_str_index sptr;
			switch(sp->access_typed()->type()->cat()) {
				case mlc::TCAT_BOOL:
					bval = sp->access_const()->access_bool_const()->bool_value();
					ctx->tab->add_const_bool(l_id, id, bval);
					break;
				case mlc::TCAT_INT:
					iptr = sp->access_const()->access_int_const()->int_value();
					ctx->tab->add_const_int(l_id, id, iptr);
					break;
				case mlc::TCAT_REAL:
					rptr = sp->access_const()->access_real_const()->real_value();
					ctx->tab->add_const_real(l_id, id, rptr);
					break;
				case mlc::TCAT_STR:
					sptr = sp->access_const()->access_str_const()->str_value();
					ctx->tab->add_const_str(l_id, id, sptr);
					break;
				case TCAT_UNDEF:
				case TCAT_ARRAY:
				case TCAT_RECORD:
				case TCAT_RANGE:
				default:
					assert(false);
					break;
			}
		}

	}

	void
	make_const_int(mlc::MlaskalCtx* ctx, mlc::ls_id_type::const_pointer id, int sign, mlc::ls_int_type::const_pointer val, int l_id) {

		switch(sign) {
			case DUTOKGE_PLUS:
				ctx->tab->add_const_int(l_id, id, val);
				break;
			case DUTOKGE_MINUS:
				mlc::ls_int_type::const_pointer nval = ctx->tab->ls_int().add( - *val);
				ctx->tab->add_const_int(l_id, id, nval);
				break;
		}
	}

	void
	make_const_real(mlc::MlaskalCtx* ctx, mlc::ls_id_type::const_pointer id, int sign, mlc::ls_real_type::const_pointer val, int l_id) {
		switch(sign) {
			case DUTOKGE_PLUS:
				ctx->tab->add_const_real(l_id, id, val);
				break;
			case DUTOKGE_MINUS:
				mlc::ls_real_type::const_pointer nval = ctx->tab->ls_real().add( - *val);
				ctx->tab->add_const_real(l_id, id, nval);
				break;
		}
	}


	mlc::ls_int_type::const_pointer
	ord_const_from_id(MlaskalCtx* ctx, int sign, mlc::ls_id_type::const_pointer id, int l_id) {

		mlc::symbol_pointer sp = ctx->tab->find_symbol(id);
		//mlc::ls_int_type::const_pointer ci;

		if ( sp->kind() != SKIND_CONST ) {
			error( DUERR_NOTCONST, l_id, *id);
		} else if(sp->access_typed()->type()->cat() != TCAT_INT) {
			error( DUERR_NOTINTEGRAL, l_id, *id);
			//TODO character const?
		} else {
			switch(sign) {
				case DUTOKGE_PLUS:
					return sp->access_const()->access_int_const()->int_value();
					break;
				case DUTOKGE_MINUS:
					return ctx->tab->ls_int().add( - *sp->access_const()->access_int_const()->int_value());
					break;
			}
		}
		return sp->access_const()->access_int_const()->int_value(); // just to shut them up
	}

	mlc::ls_int_type::const_pointer
	ord_const_from_uint(MlaskalCtx* ctx, int sign, mlc::ls_int_type::const_pointer int_ci) {
		if(sign == DUTOKGE_MINUS) {
			return ctx->tab->ls_int().add( - *int_ci);
		}

		return int_ci;
	}

	mlc::type_pointer
	range_from_id_ord_const(mlc::MlaskalCtx* ctx, mlc::ls_id_type::const_pointer id_lower, mlc::ls_int_type::const_pointer int_upper, int l_lower )  {

		mlc::ls_int_type::const_pointer int_lower = ord_const_from_id(ctx, DUTOKGE_PLUS, id_lower, l_lower);

		if(*int_lower > *int_upper) {
			error( DUERR_BADRANGE, l_lower);
		}

		// create range, even if it is invalid
		return ctx->tab->create_range_type(int_lower, int_upper);
	}

	mlc::type_pointer
	range_from_ord_const_ord_const(mlc::MlaskalCtx* ctx, mlc::ls_int_type::const_pointer int_lower, mlc::ls_int_type::const_pointer int_upper, int l_lower)  {

		if(*int_lower > *int_upper) {
			error( DUERR_BADRANGE, l_lower);
		}

		// create range, even if it is invalid
		return ctx->tab->create_range_type(int_lower, int_upper);
	}

	void
	var_from_vec(mlc::MlaskalCtx* ctx, const id_ci_vec & vars, mlc::type_pointer type) {

		for(id_ci_vec_criter rit = vars.rbegin(); rit != vars.rend(); ++rit) {
			ctx->tab->add_var(rit->second, rit->first, type);
		}
	}

	mlc::type_pointer
	get_scalar_type(MlaskalCtx* ctx, mlc::ls_id_type::const_pointer id, int l_id) {
		mlc::symbol_pointer sp = ctx->tab->find_symbol(id);
		if(sp->kind() != mlc::SKIND_TYPE) {
			error(DUERR_NOTTYPE, l_id, *id);
			return	ctx->tab->logical_undef();
		}
		switch(sp->access_typed()->type()->cat()) {
			case TCAT_BOOL:
			case TCAT_INT:
			case TCAT_REAL:
			case TCAT_STR:
			case TCAT_RANGE:
				return sp->access_type()->type();
				break;
			case TCAT_UNDEF:
			case TCAT_ARRAY:
			case TCAT_RECORD:
				error(DUERR_NOTSCALAR, l_id, *id);
				return	ctx->tab->logical_undef();
				break;
		}

		return	ctx->tab->logical_undef();
	}

	mlc::type_pointer
	get_type(MlaskalCtx* ctx, mlc::ls_id_type::const_pointer id, int l_id) {
		mlc::symbol_pointer sp = ctx->tab->find_symbol(id);
		if(sp->kind() != SKIND_TYPE) {
			error(DUERR_NOTTYPE, l_id, *id);
		}

		return sp->access_typed()->type();
	}

	void
	params_append_byref(mlc::parameter_list* params, const id_ci_vec & ids, mlc::type_pointer type) {
		for(id_ci_vec_citer it = ids.begin(); it != ids.end(); ++it) {
			params->append_parameter_by_reference(it->first, type);
		}
	}

	void
	params_append_byval(mlc::parameter_list* params, const id_ci_vec & ids, mlc::type_pointer type) {
		for(id_ci_vec_citer it = ids.begin(); it != ids.end(); ++it) {
			params->append_parameter_by_value(it->first, type);
		}
	}

	/************************************************************************************************/
	void array_deref(MlaskalExpr & expr, mlc::symbol_pointer sp) {
		mlc::icblock_pointer ip = mlc::icblock_create();

		if(sp->kind() == SKIND_GLOBAL_VARIABLE) {
			ip->append_instruction( new ai::GREF(sp->access_variable()->address()));
		} else if(sp->kind() == SKIND_LOCAL_VARIABLE) {
			ip->append_instruction( new ai::LREF(sp->access_variable()->address()));
		} else {
			assert(false);
		}

		mlc::icblock_append_delete(ip, expr.icb_p_);
		expr.icb_p_ = ip;
		expr.icb_p_->append_instruction( new ai::ADDP());

	}

	void expr_deptr(MlaskalExpr & expr) {
		if(expr.is_p_) {
			add_xldx(expr);
			expr.is_p_ = false;
		}
	}

	void expr_deref(MlaskalCtx* ctx, MlaskalExpr & expr) {
		if(expr.byref_) {

			mlc::symbol_pointer sp = ctx->tab->find_symbol(expr.id_ci_);

			if(sp->kind() == SKIND_FUNCTION) {
					expr.init_unref();
					expr_vec foo;
					call_fun(ctx, expr.icb_p_, sp->access_subprogram(), foo, expr.id_ci_, expr.id_l_);
					expr.extp_ = sp->access_typed()->type();

			} else if (sp->kind() == SKIND_GLOBAL_VARIABLE) {
				mlc::variable_symbol_pointer gv = ctx->tab->find_symbol(expr.id_ci_)->access_variable();
				if(sp->access_typed()->type()->cat() == TCAT_INT) {
					expr.init_unref();
					expr.icb_p_->append_instruction( new ai::GLDI(gv->address()));
					expr.extp_ = sp->access_typed()->type();

				} else if(sp->access_typed()->type()->cat() == TCAT_REAL) {
					expr.init_unref();
					expr.icb_p_->append_instruction( new ai::GLDR(gv->address()));
					expr.extp_ = sp->access_typed()->type();

				} else if(sp->access_typed()->type()->cat() == TCAT_BOOL) {
					expr.init_unref();
					expr.icb_p_->append_instruction( new ai::GLDB(gv->address()));
					expr.extp_ = sp->access_typed()->type();

				} else if(sp->access_typed()->type()->cat() == TCAT_STR) {
					expr.init_unref();
					expr.icb_p_->append_instruction( new ai::GLDS(gv->address()));
					expr.extp_ = sp->access_typed()->type();

				} else if(sp->access_typed()->type()->cat() == TCAT_ARRAY) {
					array_deref(expr, sp);
				} else {
					/* TODO could be other types */
					error(DUERR_CANNOTCONVERT, expr.id_l_);
					expr.init_unref();
					expr.extp_ = ctx->tab->logical_undef();
				}
				
			} else if (sp->kind() == SKIND_LOCAL_VARIABLE) {
				mlc::variable_symbol_pointer lv = ctx->tab->find_symbol(expr.id_ci_)->access_variable();
				if(sp->access_typed()->type()->cat() == TCAT_INT) {
					expr.init_unref();
					expr.icb_p_->append_instruction( new ai::LLDI(lv->address()));
					expr.extp_ = sp->access_typed()->type();

				} else if(sp->access_typed()->type()->cat() == TCAT_REAL) {
					expr.init_unref();
					expr.icb_p_->append_instruction( new ai::LLDR(lv->address()));
					expr.extp_ = sp->access_typed()->type();

				} else if(sp->access_typed()->type()->cat() == TCAT_BOOL) {
					expr.init_unref();
					expr.icb_p_->append_instruction( new ai::LLDB(lv->address()));
					expr.extp_ = sp->access_typed()->type();

				} else if(sp->access_typed()->type()->cat() == TCAT_STR) {
					expr.init_unref();
					expr.icb_p_->append_instruction( new ai::LLDS(lv->address()));
					expr.extp_ = sp->access_typed()->type();

				} else if(sp->access_typed()->type()->cat() == TCAT_ARRAY) {
					array_deref(expr, sp);
				} else {
					/* TODO could be other types */
					error(DUERR_CANNOTCONVERT, expr.id_l_); 
					expr.init_unref();
					expr.extp_ = ctx->tab->logical_undef();
				}

			} else if (sp->kind() == SKIND_CONST) {
				if(sp->access_typed()->type()->cat() == TCAT_INT) {
					expr.init_unref();
					mlc::int_const_symbol_pointer ic = ctx->tab->find_symbol(expr.id_ci_)->access_const()->access_int_const();
					expr.icb_p_->append_instruction( new ai::LDLITI(ic->int_value()));
					expr.extp_ = sp->access_typed()->type();

				} else if(sp->access_typed()->type()->cat() == TCAT_REAL) {
					expr.init_unref();
					mlc::real_const_symbol_pointer rc = ctx->tab->find_symbol(expr.id_ci_)->access_const()->access_real_const();
					expr.icb_p_->append_instruction( new ai::LDLITR(rc->real_value()));
					expr.extp_ = sp->access_typed()->type();

				} else if(sp->access_typed()->type()->cat() == TCAT_BOOL) {
					expr.init_unref();
					mlc::bool_const_symbol_pointer bc = ctx->tab->find_symbol(expr.id_ci_)->access_const()->access_bool_const();
					expr.icb_p_->append_instruction( new ai::LDLITB(bc->bool_value()));
					expr.extp_ = sp->access_typed()->type();

				} else if(sp->access_typed()->type()->cat() == TCAT_STR) {
					expr.init_unref();
					mlc::str_const_symbol_pointer sc = ctx->tab->find_symbol(expr.id_ci_)->access_const()->access_str_const();
					expr.icb_p_->append_instruction( new ai::LDLITS(sc->str_value()));
					expr.extp_ = sp->access_typed()->type();

				} else {
					error(DUERR_CANNOTCONVERT, expr.id_l_);
					expr.init_unref();
					expr.extp_ = ctx->tab->logical_undef();
				}

			} else {
				error(DUERR_NOTVAR, expr.id_l_, *(expr.id_ci_));
				expr.init_unref();
				expr.extp_ = ctx->tab->logical_undef();
			}

		}

		//expr_deptr(expr);
	}

	void
	add_gstx(MlaskalCtx* ctx, MlaskalExpr & expr, mlc::ls_id_type::const_pointer id, int l_id) {

		mlc::global_variable_symbol_pointer gv = ctx->tab->find_symbol(id)->access_global_variable();

		switch(expr.extp_->cat()) {
			case TCAT_BOOL:
				if(gv->access_typed()->type()->cat() == TCAT_BOOL) {
					expr.icb_p_->append_instruction(new ai::GSTB(gv->address()));
				} else {
					error(DUERR_CANNOTCONVERT, l_id); 
				}
				break;
			case TCAT_INT:
				/* expression is of integer type, store to: */
				if(gv->access_typed()->type()->cat() == TCAT_INT) {
					expr.icb_p_->append_instruction(new ai::GSTI(gv->address()));
				} else if(gv->access_typed()->type()->cat() == TCAT_REAL) {
					/* silent convert */
					expr.icb_p_->append_instruction(new ai::CVRTIR());
					expr.icb_p_->append_instruction(new ai::GSTR(gv->address()));
				} else {
					error(DUERR_CANNOTCONVERT, l_id); 
				}
				break;
			case TCAT_REAL:
				/* expression is of real type, store to: */
				if(gv->access_typed()->type()->cat() == TCAT_REAL) {
					expr.icb_p_->append_instruction(new ai::GSTR(gv->address()));
				} else if(gv->access_typed()->type()->cat() == TCAT_INT) {
					/* verbose convert */
					expr.icb_p_->append_instruction(new ai::CVRTRI());
					expr.icb_p_->append_instruction(new ai::GSTI(gv->address()));
					error(DUERR_CONVERSION, l_id);
				} else {
					error(DUERR_CANNOTCONVERT, l_id); 
				}
				break;
			case TCAT_STR:
				if(gv->access_typed()->type()->cat() == TCAT_STR) {
					expr.icb_p_->append_instruction(new ai::GSTS(gv->address()));
				} else {
					error(DUERR_CANNOTCONVERT, l_id); 
				}
				break;
			case TCAT_RANGE:
					/* forever undef */
					assert(false);
				break;
			case TCAT_ARRAY:
					/* sofar undef */
					assert(false);
				break;
			case TCAT_RECORD:
					/* forever undef */
					assert(false);
				break;
			case TCAT_UNDEF:
					/* wtf? */
					assert(false);
				break;
		}
	}

	void
	add_lstx(MlaskalCtx* ctx, MlaskalExpr & expr, mlc::ls_id_type::const_pointer id, int l_id) {

		mlc::local_variable_symbol_pointer lv = ctx->tab->find_symbol(id)->access_local_variable();

		switch(expr.extp_->cat()) {
			case TCAT_BOOL:
				if(lv->access_typed()->type()->cat() == TCAT_BOOL) {
					expr.icb_p_->append_instruction(new ai::LSTB(lv->address()));
				} else {
					error(DUERR_CANNOTCONVERT, l_id); 
				}
				break;
			case TCAT_INT:
				/* expression is of integer type, store to: */
				if(lv->access_typed()->type()->cat() == TCAT_INT) {
					expr.icb_p_->append_instruction(new ai::LSTI(lv->address()));
				} else if(lv->access_typed()->type()->cat() == TCAT_REAL) {
					/* silent convert */
					expr.icb_p_->append_instruction(new ai::CVRTIR());
					expr.icb_p_->append_instruction(new ai::LSTR(lv->address()));
				} else {
					error(DUERR_CANNOTCONVERT, l_id); 
				}
				break;
			case TCAT_REAL:
				/* expression is of real type, store to: */
				if(lv->access_typed()->type()->cat() == TCAT_REAL) {
					expr.icb_p_->append_instruction(new ai::LSTR(lv->address()));
				} else if(lv->access_typed()->type()->cat() == TCAT_INT) {
					/* verbose convert */
					expr.icb_p_->append_instruction(new ai::CVRTRI());
					expr.icb_p_->append_instruction(new ai::LSTI(lv->address()));
					error(DUERR_CONVERSION, l_id);
				} else {
					error(DUERR_CANNOTCONVERT, l_id); 
				}
				break;
			case TCAT_STR:
				if(lv->access_typed()->type()->cat() == TCAT_STR) {
					expr.icb_p_->append_instruction(new ai::LSTS(lv->address()));
				} else {
					error(DUERR_CANNOTCONVERT, l_id); 
				}
				break;
			case TCAT_RANGE:
					/* sofar undef */
					assert(false);
				break;
			case TCAT_ARRAY:
					/* sofar undef */
					assert(false);
				break;
			case TCAT_RECORD:
					/* forever undef */
					assert(false);
				break;
			case TCAT_UNDEF:
					/* wtf? */
					assert(false);
				break;
		}

	}


	void
	add_fncret(MlaskalCtx* ctx, MlaskalExpr & expr, mlc::ls_id_type::const_pointer id, int l_id) {

		mlc::function_symbol_pointer fn = ctx->tab->find_symbol(id)->access_function();

		switch(expr.extp_->cat()) {
			case TCAT_BOOL:
				if(fn->access_typed()->type()->cat() == TCAT_BOOL) {
					expr.icb_p_->append_instruction(new ai::LSTB(ctx->tab->my_return_address()));
				} else {
					error(DUERR_CANNOTCONVERT, l_id); 
				}
				break;
			case TCAT_INT:
				/* expression is of integer type, store to: */
				if(fn->access_typed()->type()->cat() == TCAT_INT) {
					expr.icb_p_->append_instruction(new ai::LSTI(ctx->tab->my_return_address()));
				} else if(fn->access_typed()->type()->cat() == TCAT_REAL) {
					/* silent convert */
					expr.icb_p_->append_instruction(new ai::CVRTIR());
					expr.icb_p_->append_instruction(new ai::LSTR(ctx->tab->my_return_address()));
				} else {
					error(DUERR_CANNOTCONVERT, l_id); 
				}
				break;
			case TCAT_REAL:
				/* expression is of real type, store to: */
				if(fn->access_typed()->type()->cat() == TCAT_REAL) {
					expr.icb_p_->append_instruction(new ai::LSTR(ctx->tab->my_return_address()));
				} else if(fn->access_typed()->type()->cat() == TCAT_INT) {
					/* verbose convert */
					expr.icb_p_->append_instruction(new ai::CVRTRI());
					expr.icb_p_->append_instruction(new ai::LSTI(ctx->tab->my_return_address()));
					error(DUERR_CONVERSION, l_id);
				} else {
					error(DUERR_CANNOTCONVERT, l_id); 
				}
				break;
			case TCAT_STR:
				if(fn->access_typed()->type()->cat() == TCAT_STR) {
					expr.icb_p_->append_instruction(new ai::LSTS(ctx->tab->my_return_address()));
				} else {
					error(DUERR_CANNOTCONVERT, l_id); 
				}
				break;
			case TCAT_RANGE:
					/* sofar undef */
					assert(false);
				break;
			case TCAT_ARRAY:
					/* sofar undef */
					assert(false);
				break;
			case TCAT_RECORD:
					/* forever undef */
					assert(false);
				break;
			case TCAT_UNDEF:
					/* wtf? */
					assert(false);
				break;
		}

	}

	/* TODO by reference */
	void
	call_proc(MlaskalCtx* ctx, mlc::icblock_pointer icb_p, mlc::subprogram_symbol_pointer proc, expr_vec rpar, mlc::ls_id_type::const_pointer id, int l_id) {
		
		expr_vec::reverse_iterator rit = rpar.rbegin();
		mlc::parameter_list::const_iterator cit = proc->parameters()->begin();

		mlc::icblock_pointer foo;
		mlc::icblock_pointer dtrblc = mlc::icblock_create();


		for( ; (cit != proc->parameters()->end()) && (rit != rpar.rend()); ++cit, ++rit) {
			/* TODO by ref */

			expr_deref(ctx, *rit);
			expr_deptr(*rit);

			mlc::icblock_append_delete(icb_p, rit->icb_p_);
			if(!mlc::identical_type(cit->ltype, rit->extp_)) {

				if((cit->ltype->cat() == mlc::TCAT_REAL) && rit->extp_->cat() == mlc::TCAT_INT) {
					/* silent convert */
					icb_p->append_instruction( new ai::CVRTIR());

				} else if((cit->ltype->cat() == mlc::TCAT_INT) && (rit->extp_->cat() == mlc::TCAT_REAL)) {
					/* verbose convert */
					icb_p->append_instruction( new ai::CVRTRI());
					error(DUERR_CONVERSION, l_id);

				} else {
					error(DUERR_CANNOTCONVERT, l_id);
				}

			}

			if(cit->ltype->cat() == mlc::TCAT_INT) {
				foo = mlc::icblock_create();
				foo->append_instruction( new ai::DTORI());
				mlc::icblock_append_delete(foo, dtrblc);
				dtrblc = foo;
			} else if(cit->ltype->cat() == mlc::TCAT_REAL) {
				foo = mlc::icblock_create();
				foo->append_instruction( new ai::DTORR());
				mlc::icblock_append_delete(foo, dtrblc);
				dtrblc = foo;
			} else if(cit->ltype->cat() == mlc::TCAT_BOOL) {
				foo = mlc::icblock_create();
				foo->append_instruction( new ai::DTORB());
				mlc::icblock_append_delete(foo, dtrblc);
				dtrblc = foo;
			} else if(cit->ltype->cat() == mlc::TCAT_STR) {
				foo = mlc::icblock_create();
				foo->append_instruction( new ai::DTORS());
				mlc::icblock_append_delete(foo, dtrblc);
				dtrblc = foo;
			} else {
				assert(false);
			}

		}

		if((cit != proc->parameters()->end()) || (rit != rpar.rend())) {
			error(DUERR_PARNUM, l_id, *id);
		}
		
		icb_p->append_instruction( new ai::CALL(proc->code()));
		mlc::icblock_append_delete(icb_p, dtrblc);

	}

	void
	call_fun(MlaskalCtx* ctx, mlc::icblock_pointer icb_p, mlc::subprogram_symbol_pointer fun, expr_vec rpar, mlc::ls_id_type::const_pointer id, int l_id, bool destr_ret) {
	
		mlc::icblock_pointer dtrret = mlc::icblock_create();

		if(fun->access_function()->type()->cat() == TCAT_INT) {
			icb_p->append_instruction( new ai::INITI());
			if(destr_ret)
				dtrret->append_instruction( new ai::DTORI());

		} else if(fun->access_function()->type()->cat() == TCAT_REAL) {
			icb_p->append_instruction( new ai::INITB());
			if(destr_ret)
				dtrret->append_instruction( new ai::DTORR());

		} else if(fun->access_function()->type()->cat() == TCAT_BOOL) {
			icb_p->append_instruction( new ai::INITR());
			if(destr_ret)
				dtrret->append_instruction( new ai::DTORB());

		} else if(fun->access_function()->type()->cat() == TCAT_STR) {
			icb_p->append_instruction( new ai::INITS());
			if(destr_ret)
				dtrret->append_instruction( new ai::DTORS());
		} else if(fun->access_function()->type()->cat() == TCAT_RANGE) {
			/* TODO */
		} else {
			assert(false);
		}

		/* tiny fake */
		call_proc(ctx, icb_p, fun, rpar, id, l_id);

		mlc::icblock_append_delete(icb_p, dtrret);	
	}

	void
	expr_add( MlaskalCtx* ctx, MlaskalExpr & res, MlaskalExpr & lop, MlaskalExpr & rop, int l_op) {
		expr_deref(ctx, lop);
		expr_deref(ctx, rop);
		expr_deptr(lop);
		expr_deptr(rop);
		res.is_p_ = false;

		if(lop.extp_->cat() == TCAT_INT) {
			if(rop.extp_->cat() == TCAT_INT) {
				mlc::icblock_append_delete(res.icb_p_, lop.icb_p_);
				mlc::icblock_append_delete(res.icb_p_, rop.icb_p_);
				res.icb_p_->append_instruction( new ai::ADDI());
				res.extp_ = lop.extp_;

			} else if(rop.extp_->cat() == TCAT_REAL) {
				/* silent convert to real */
				mlc::icblock_append_delete(res.icb_p_, lop.icb_p_);
				res.icb_p_->append_instruction( new ai::CVRTIR);
				mlc::icblock_append_delete(res.icb_p_, rop.icb_p_);
				res.icb_p_->append_instruction( new ai::ADDR());
				res.extp_ = rop.extp_;
				
			} else {
				error(DUERR_CANNOTCONVERT, l_op);
			}
		} else if(lop.extp_->cat() == TCAT_REAL) {
			if(rop.extp_->cat() == TCAT_INT) {
				mlc::icblock_append_delete(res.icb_p_, lop.icb_p_);
				/* silent convert to real */
				mlc::icblock_append_delete(res.icb_p_, rop.icb_p_);
				res.icb_p_->append_instruction( new ai::CVRTIR);
				res.icb_p_->append_instruction( new ai::ADDR());
				res.extp_ = lop.extp_;

			} else if(rop.extp_->cat() == TCAT_REAL) {
				mlc::icblock_append_delete(res.icb_p_, lop.icb_p_);
				mlc::icblock_append_delete(res.icb_p_, rop.icb_p_);
				res.icb_p_->append_instruction( new ai::ADDR());
				res.extp_ = rop.extp_;
				
			} else {
				error(DUERR_CANNOTCONVERT, l_op);
			}
		} else if(lop.extp_->cat() == TCAT_STR) {
			if(rop.extp_->cat() == TCAT_STR) {
				mlc::icblock_append_delete(res.icb_p_, lop.icb_p_);
				mlc::icblock_append_delete(res.icb_p_, rop.icb_p_);
				res.icb_p_->append_instruction( new ai::ADDS());
				res.extp_ = lop.extp_;

			} else {
				error(DUERR_CANNOTCONVERT, l_op);
			}
		} else {
			error(DUERR_CANNOTCONVERT, l_op);

		}
	}

	void
	expr_mul( MlaskalCtx* ctx, MlaskalExpr & res, MlaskalExpr & lop, MlaskalExpr & rop, int l_op) {
		expr_deref(ctx, lop);
		expr_deref(ctx, rop);
		expr_deptr(lop);
		expr_deptr(rop);
		res.is_p_ = false;

		if(lop.extp_->cat() == TCAT_INT) {
			if(rop.extp_->cat() == TCAT_INT) {
				mlc::icblock_append_delete(res.icb_p_, lop.icb_p_);
				mlc::icblock_append_delete(res.icb_p_, rop.icb_p_);
				res.icb_p_->append_instruction( new ai::MULI());

				res.extp_ = lop.extp_;

			} else if(rop.extp_->cat() == TCAT_REAL) {
				/* silent convert to real */
				mlc::icblock_append_delete(res.icb_p_, lop.icb_p_);
				res.icb_p_->append_instruction( new ai::CVRTIR);
				mlc::icblock_append_delete(res.icb_p_, rop.icb_p_);
				res.icb_p_->append_instruction( new ai::MULR());
				res.extp_ = rop.extp_;
				
			} else {
				error(DUERR_CANNOTCONVERT, l_op);
			}
		} else if(lop.extp_->cat() == TCAT_REAL) {
			if(rop.extp_->cat() == TCAT_INT) {
				mlc::icblock_append_delete(res.icb_p_, lop.icb_p_);
				/* silent convert to real */
				mlc::icblock_append_delete(res.icb_p_, rop.icb_p_);
				res.icb_p_->append_instruction( new ai::CVRTIR);
				res.icb_p_->append_instruction( new ai::MULR());
				res.extp_ = lop.extp_;

			} else if(rop.extp_->cat() == TCAT_REAL) {
				mlc::icblock_append_delete(res.icb_p_, lop.icb_p_);
				mlc::icblock_append_delete(res.icb_p_, rop.icb_p_);
				res.icb_p_->append_instruction( new ai::MULR());
				res.extp_ = rop.extp_;
				
			} else {
				error(DUERR_CANNOTCONVERT, l_op);
			}
		} else {
			error(DUERR_CANNOTCONVERT, l_op);

		}


	}

	void
	expr_sub( MlaskalCtx* ctx, MlaskalExpr & res, MlaskalExpr & lop, MlaskalExpr & rop, int l_op) {
		expr_deref(ctx, lop);
		expr_deref(ctx, rop);
		expr_deptr(lop);
		expr_deptr(rop);
		res.is_p_ = false;

		if(lop.extp_->cat() == TCAT_INT) {
			if(rop.extp_->cat() == TCAT_INT) {
				mlc::icblock_append_delete(res.icb_p_, lop.icb_p_);
				mlc::icblock_append_delete(res.icb_p_, rop.icb_p_);
				res.icb_p_->append_instruction( new ai::SUBI());

				res.extp_ = lop.extp_;

			} else if(rop.extp_->cat() == TCAT_REAL) {
				/* silent convert to real */
				mlc::icblock_append_delete(res.icb_p_, lop.icb_p_);
				res.icb_p_->append_instruction( new ai::CVRTIR);
				mlc::icblock_append_delete(res.icb_p_, rop.icb_p_);
				res.icb_p_->append_instruction( new ai::SUBR());
				res.extp_ = rop.extp_;
				
			} else {
				error(DUERR_CANNOTCONVERT, l_op);
			}
		} else if(lop.extp_->cat() == TCAT_REAL) {
			if(rop.extp_->cat() == TCAT_INT) {
				mlc::icblock_append_delete(res.icb_p_, lop.icb_p_);
				/* silent convert to real */
				mlc::icblock_append_delete(res.icb_p_, rop.icb_p_);
				res.icb_p_->append_instruction( new ai::CVRTIR);
				res.icb_p_->append_instruction( new ai::SUBR());
				res.extp_ = lop.extp_;

			} else if(rop.extp_->cat() == TCAT_REAL) {
				mlc::icblock_append_delete(res.icb_p_, lop.icb_p_);
				mlc::icblock_append_delete(res.icb_p_, rop.icb_p_);
				res.icb_p_->append_instruction( new ai::SUBR());
				res.extp_ = rop.extp_;
				
			} else {
				error(DUERR_CANNOTCONVERT, l_op);
			}
		} else {
			error(DUERR_CANNOTCONVERT, l_op);

		}


	}

	void
	expr_divr( MlaskalCtx* ctx, MlaskalExpr & res, MlaskalExpr & lop, MlaskalExpr & rop, int l_op) {
		expr_deref(ctx, lop);
		expr_deref(ctx, rop);
		expr_deptr(lop);
		expr_deptr(rop);
		res.is_p_ = false;

		if(lop.extp_->cat() == TCAT_INT) {
			if(rop.extp_->cat() == TCAT_INT) {
				mlc::icblock_append_delete(res.icb_p_, lop.icb_p_);
				res.icb_p_->append_instruction( new ai::CVRTIR);
				mlc::icblock_append_delete(res.icb_p_, rop.icb_p_);
				res.icb_p_->append_instruction( new ai::CVRTIR);
				res.icb_p_->append_instruction( new ai::DIVR());

				res.extp_ = ctx->tab->logical_real();

			} else if(rop.extp_->cat() == TCAT_REAL) {
				mlc::icblock_append_delete(res.icb_p_, lop.icb_p_);
				res.icb_p_->append_instruction( new ai::CVRTIR);
				mlc::icblock_append_delete(res.icb_p_, rop.icb_p_);
				res.icb_p_->append_instruction( new ai::DIVR());
				res.extp_ = rop.extp_;
				
			} else {
				error(DUERR_CANNOTCONVERT, l_op);
			}
		} else if(lop.extp_->cat() == TCAT_REAL) {
			if(rop.extp_->cat() == TCAT_INT) {
				mlc::icblock_append_delete(res.icb_p_, lop.icb_p_);
				/* silent convert to real */
				mlc::icblock_append_delete(res.icb_p_, rop.icb_p_);
				res.icb_p_->append_instruction( new ai::CVRTIR);
				res.icb_p_->append_instruction( new ai::DIVR());
				res.extp_ = lop.extp_;

			} else if(rop.extp_->cat() == TCAT_REAL) {
				mlc::icblock_append_delete(res.icb_p_, lop.icb_p_);
				mlc::icblock_append_delete(res.icb_p_, rop.icb_p_);
				res.icb_p_->append_instruction( new ai::DIVR());
				res.extp_ = rop.extp_;
				
			} else {
				error(DUERR_CANNOTCONVERT, l_op);
			}
		} else {
			error(DUERR_CANNOTCONVERT, l_op);

		}
	}

	void
	expr_divi( MlaskalCtx* ctx, MlaskalExpr & res, MlaskalExpr & lop, MlaskalExpr & rop, int l_op) {
		expr_deref(ctx, lop);
		expr_deref(ctx, rop);
		expr_deptr(lop);
		expr_deptr(rop);
		res.is_p_ = false;

		if(lop.extp_->cat() == TCAT_INT) {
			if(rop.extp_->cat() == TCAT_INT) {
				mlc::icblock_append_delete(res.icb_p_, lop.icb_p_);
				mlc::icblock_append_delete(res.icb_p_, rop.icb_p_);
				res.icb_p_->append_instruction( new ai::DIVI());

				res.extp_ = rop.extp_;
			} else {
				error(DUERR_CANNOTCONVERT, l_op);
			}
		} else {
			error(DUERR_CANNOTCONVERT, l_op);

		}
	}

	void
	expr_mod( MlaskalCtx* ctx, MlaskalExpr & res, MlaskalExpr & lop, MlaskalExpr & rop, int l_op) {
		expr_deref(ctx, lop);
		expr_deref(ctx, rop);
		expr_deptr(lop);
		expr_deptr(rop);
		res.is_p_ = false;

		if(lop.extp_->cat() == TCAT_INT) {
			if(rop.extp_->cat() == TCAT_INT) {
				mlc::icblock_append_delete(res.icb_p_, lop.icb_p_);
				mlc::icblock_append_delete(res.icb_p_, rop.icb_p_);
				res.icb_p_->append_instruction( new ai::MODI());

				res.extp_ = rop.extp_;
			} else {
				error(DUERR_CANNOTCONVERT, l_op);
			}
		} else {
			error(DUERR_CANNOTCONVERT, l_op);

		}
	}

	void
	add_xstx(MlaskalCtx* ctx, mlc::icblock_pointer icb_p, MlaskalExpr & arr, MlaskalExpr & expr) {
		expr_deref(ctx, arr);
		expr_deref(ctx, expr);


		if(arr.extp_->cat() == mlc::TCAT_INT) {
			/* loads value on stack */
			if(expr.extp_->cat() == TCAT_INT) {
				mlc::icblock_append_delete(icb_p, expr.icb_p_);
			} else if(expr.extp_->cat() == TCAT_REAL) {
				mlc::icblock_append_delete(icb_p, expr.icb_p_);
				icb_p->append_instruction( new ai::CVRTRI());
				error(DUERR_CONVERSION, arr.id_l_);
			} else {
				error(DUERR_CANNOTCONVERT, arr.id_l_);
			}

			/* loads pointer into array on stack */
			mlc::icblock_append_delete(icb_p, arr.icb_p_);

			icb_p->append_instruction( new ai::XSTI());
		} else if(arr.extp_->cat() == mlc::TCAT_REAL) {
			/* loads value on stack */
			if(expr.extp_->cat() == TCAT_INT) {
				mlc::icblock_append_delete(icb_p, expr.icb_p_);
				icb_p->append_instruction( new ai::CVRTIR());
			} else if(expr.extp_->cat() == TCAT_REAL) {
				mlc::icblock_append_delete(icb_p, expr.icb_p_);
			} else {
				error(DUERR_CANNOTCONVERT, arr.id_l_);
			}

			/* loads pointer into array on stack */
			mlc::icblock_append_delete(icb_p, arr.icb_p_);

			icb_p->append_instruction( new ai::XSTR());
		} else if((arr.extp_->cat() == mlc::TCAT_BOOL) && (expr.extp_->cat() == mlc::TCAT_BOOL)) {
			/* loads value on stack */
			mlc::icblock_append_delete(icb_p, expr.icb_p_);

			/* loads pointer into array on stack */
			mlc::icblock_append_delete(icb_p, arr.icb_p_);

			icb_p->append_instruction( new ai::XSTB());
		} else if((arr.extp_->cat() == mlc::TCAT_STR) && (expr.extp_->cat() == mlc::TCAT_STR)) {
			/* loads value on stack */
			mlc::icblock_append_delete(icb_p, expr.icb_p_);

			/* loads pointer into array on stack */
			mlc::icblock_append_delete(icb_p, arr.icb_p_);

			icb_p->append_instruction( new ai::XSTS());
		} else {
			//error(DUERR_CANNOTCONVERT, 
			assert(false);
		}
	}

	void
	add_xldx(MlaskalExpr & expr) {

		if(expr.extp_->cat() == TCAT_INT) {
			expr.icb_p_->append_instruction( new ai::XLDI());
		} else if(expr.extp_->cat() == TCAT_REAL) {
			expr.icb_p_->append_instruction( new ai::XLDR());
		} else if(expr.extp_->cat() == TCAT_BOOL) {
			expr.icb_p_->append_instruction( new ai::XLDB());
		} else if(expr.extp_->cat() == TCAT_STR) {
			expr.icb_p_->append_instruction( new ai::XLDS());
		} else {
			assert(false);
		}
	}

	void
	add_minus(MlaskalExpr & expr, int l) {

		if(expr.extp_->cat() == TCAT_INT) {
			expr.icb_p_->append_instruction( new ai::MINUSI());
		} else if(expr.extp_->cat() == TCAT_REAL) {
			expr.icb_p_->append_instruction( new ai::MINUSR());
		} else {
			error( DUERR_CANNOTCONVERT, l);
		}

	}
	
	void
	arr_index(MlaskalCtx* ctx, MlaskalExpr & expr, expr_vec & indicies, bool appending) {

		if(expr.extp_->cat() != TCAT_ARRAY) {
			assert(false); 
			//TODO not indexing an array
		}

		mlc::array_type_pointer arr = expr.extp_->access_array();
	
		////

		bool first = true;
		for(expr_vec::reverse_iterator rit = indicies.rbegin(); rit != indicies.rend(); ++rit) {
			if(!first) {
				//recursion
				if(arr->element_type()->cat() == TCAT_ARRAY) {
				arr = arr->element_type()->access_array();
				} else {
					//TODO error, bad indexing
					assert(false);
				}
			}
								
			expr_deref(ctx, *rit);
			expr_deptr(*rit);
			if(rit->extp_->cat() != TCAT_INT) {
				error( DUERR_NOTORDINAL, expr.id_l_);
			}
			
			mlc::icblock_append_delete( expr.icb_p_, rit->icb_p_);
			if(arr->index_type()->cat() != ctx->tab->logical_integer()->cat()) {
				/* thus range */
				expr.icb_p_->append_instruction( new ai::LDLITI(arr->index_type()->access_range()->lowerBound()));
				expr.icb_p_->append_instruction( new ai::SUBI());
			}
				//correct start
			
			expr.icb_p_->append_instruction( new ai::LDLITI( ctx->tab->ls_int().add( arr->compute_size())));
			expr.icb_p_->append_instruction( new ai::MULI());
			if(!first || appending) {
				expr.icb_p_->append_instruction( new ai::ADDI());
			}
			first = false;
		}
		expr.extp_ = arr->element_type();

	}
}
