%{
	/*
	*  frantisek farka
	*/
	#include "du6.h"
	#include "du6lval.hpp"
	#include "duerr.h"
	#include "du6tok.h"
	#include "du6sem.h"


	/* semnatic value */	
	#define YYSTYPE mlc::MlaskalLval

	/* #locations */
	#define YYLTYPE unsigned
	#define YYLLOC_DEFAULT(cur, rhs, n)	do { if(n) (cur)=(rhs)[1]; else (cur)=(rhs)[0]; } while(0)

    /* local stuff */
    using namespace mlc;
    
	/* error stuff */
    #define YYERROR_VERBOSE 1
    static void yyerror(YYLTYPE *line, mlc::MlaskalCtx *ctx, const char *m);
    
%}

%token DUTOK_PROGRAM			    /* program */
%token DUTOK_LABEL			    /* label */
%token DUTOK_CONST			    /* const */
%token DUTOK_TYPE			    /* type */
%token DUTOK_VAR			    /* var */
%token DUTOK_BEGIN			    /* begin */
%token DUTOK_END			    /* end */
%token DUTOK_PROCEDURE			    /* procedure */
%token DUTOK_FUNCTION			    /* function */
%token DUTOK_ARRAY			    /* array */
%token DUTOK_OF				    /* of */
%token DUTOK_GOTO			    /* goto */
%token DUTOK_IF				    /* if */
%token DUTOK_THEN			    /* then */
%token DUTOK_ELSE			    /* else */
%token DUTOK_WHILE			    /* while */
%token DUTOK_DO				    /* do */
%token DUTOK_REPEAT			    /* repeat */
%token DUTOK_UNTIL			    /* until */
%token DUTOK_FOR			    /* for */
%token DUTOK_OR				    /* or */
%token DUTOK_NOT			    /* not */
%token DUTOK_RECORD			    /* record */

/* literals */
%token<id_ci_> DUTOK_IDENTIFIER			    /* identifier */
%token<int_ci_> DUTOK_UINT			    /* unsigned integer */
%token<real_ci_> DUTOK_REAL			    /* real number */
%token<str_ci_> DUTOK_STRING			    /* string */

/* delimiters */
%token DUTOK_SEMICOLON			    /* ; */
%token DUTOK_DOT			    /* . */
%token DUTOK_COMMA			    /* , */
%token DUTOK_EQ				    /* = */
%token DUTOK_COLON			    /* : */
%token DUTOK_LPAR			    /* ( */
%token DUTOK_RPAR			    /* ) */
%token DUTOK_DOTDOT			    /* .. */
%token DUTOK_LSBRA			    /* [ */
%token DUTOK_RSBRA			    /* ] */
%token DUTOK_ASSIGN			    /* := */

/* grouped operators and keywords */
%token<dtge_> DUTOK_OPER_REL			    /* <, <=, <>, >=, > */
%token<dtge_> DUTOK_OPER_SIGNADD		    /* +, - */
%token<dtge_> DUTOK_OPER_MUL			    /* *, /, div, mod, and */
%token<dtge_> DUTOK_FOR_DIRECTION		    /* to, downto */

%type<tp_> type_val ord_type_noid ord_type_id struc_type_val
%type<tp_vec_> ord_type_seq
%type<dtge_> sign_add_opt
%type<int_ci_> ord_const_id ord_const_noid
%type<id_ci_> procedure_header function_header
%type<id_ci_vec_> identifier_seq
%type<pl_p_> formal_params
%type<icb_p_> program_block block stmt_seq matched unmatched stmt label_inst_opt matched_seq unmatched_seq
%type<expr_> expr simple_expr term faktor variable
%type<expr_vec_> real_params ordinal_expr_seq

%start mlaskal
%error-verbose
%pure-parser
%parse-param {mlc::MlaskalCtx *ctx}
%lex-param {mlc::MlaskalCtx *ctx}
%locations

%%

mlaskal:	  DUTOK_PROGRAM DUTOK_IDENTIFIER DUTOK_SEMICOLON program_block DUTOK_DOT	{
				ctx->tab->set_main_code($2, $4);
			} 
		;

program_block:	  label_block const_block type_block var_block method_block DUTOK_BEGIN stmt_seq DUTOK_END	{
				$$ = $7;
			}
		;

label_block:	  DUTOK_LABEL uint_seq DUTOK_SEMICOLON  
		| /*no label block */
		;

const_block:	  DUTOK_CONST const_seq
		| /*no const block */
		;

type_block:	  DUTOK_TYPE type_seq
		| /*no type block */
		;

var_block:	  DUTOK_VAR var_seq 
		| /*no var block */
		;

method_block:	  procedure_header DUTOK_SEMICOLON block DUTOK_SEMICOLON method_block	{
				ctx->tab->set_subprogram_code($1, $3);
			}
		| function_header DUTOK_SEMICOLON block DUTOK_SEMICOLON method_block	{
				ctx->tab->set_subprogram_code($1, $3);
			}
		| /*no method block */
		;

block:		  label_block const_block type_block var_block DUTOK_BEGIN stmt_seq DUTOK_END {
				ctx->tab->leave(@7);
				$$ = $6;
			}
		;

procedure_header: DUTOK_PROCEDURE DUTOK_IDENTIFIER DUTOK_LPAR formal_params DUTOK_RPAR	{
				ctx->tab->add_proc(@2, $2, $4); // add_proc deallocates list
				ctx->tab->enter(@2, $2);
				$$ = $2;
			}
		| DUTOK_PROCEDURE DUTOK_IDENTIFIER	{
				ctx->tab->add_proc(@2, $2, mlc::create_parameter_list()); // add_proc deallocates list
				ctx->tab->enter(@2, $2);
				$$ = $2;
			}
		;

function_header:  DUTOK_FUNCTION DUTOK_IDENTIFIER DUTOK_LPAR formal_params DUTOK_RPAR DUTOK_COLON DUTOK_IDENTIFIER	{ /* scalar type identifier */
				ctx->tab->add_fnc(@2, $2, get_scalar_type(ctx, $7, @7), $4); // add_fun deallocates list
				ctx->tab->enter(@2, $2);
				$$ = $2;
			}
		| DUTOK_FUNCTION DUTOK_IDENTIFIER DUTOK_COLON DUTOK_IDENTIFIER	{ /* scalar type identifier */
				ctx->tab->add_fnc(@2, $2, get_scalar_type(ctx, $4, @4), mlc::create_parameter_list()); // add_fun deallocates list
				ctx->tab->enter(@2, $2);
				$$ = $2;
			}
		;

formal_params:	  DUTOK_VAR identifier_seq DUTOK_COLON DUTOK_IDENTIFIER	{ /* by ref */
				$$ = mlc::create_parameter_list();
				params_append_byref($$, $2, get_type(ctx, $4, @4));
			}
		| identifier_seq DUTOK_COLON DUTOK_IDENTIFIER	{ /* by val */
				$$ = mlc::create_parameter_list();
				params_append_byval($$, $1, get_type(ctx, $3, @3));
			}
		| formal_params DUTOK_SEMICOLON DUTOK_VAR identifier_seq DUTOK_COLON DUTOK_IDENTIFIER	{ /* by ref */
				$$ = $1;
				params_append_byref($$, $4, get_type(ctx, $6, @6));
			}
		| formal_params DUTOK_SEMICOLON identifier_seq DUTOK_COLON DUTOK_IDENTIFIER { /* by val */
				$$ = $1;
				params_append_byval($$, $3, get_type(ctx, $5, @5));
			}
		;

/* various sequences */

uint_seq:	  DUTOK_UINT	{
				ctx->tab->add_label_entry(@1, $1, mlc::new_label( ctx)); /* unique for labes, thus here */
			}
		| DUTOK_UINT DUTOK_COMMA uint_seq	{
				ctx->tab->add_label_entry(@1, $1, mlc::new_label( ctx)); /* dtto */
			}
		;

const_seq:	  DUTOK_IDENTIFIER DUTOK_EQ sign_add_opt DUTOK_UINT DUTOK_SEMICOLON {
				mlc::make_const_int(ctx, $1, $3, $4, @1);
			}
		| DUTOK_IDENTIFIER DUTOK_EQ sign_add_opt DUTOK_REAL DUTOK_SEMICOLON {
				mlc::make_const_real(ctx, $1, $3, $4, @1);
			}
		| DUTOK_IDENTIFIER DUTOK_EQ DUTOK_STRING DUTOK_SEMICOLON {
			  	ctx->tab->add_const_str(@1, $1, $3);
		  	}
		| DUTOK_IDENTIFIER DUTOK_EQ DUTOK_IDENTIFIER DUTOK_SEMICOLON /* of konst */ {
				mlc::make_const_from_id(ctx, $1, $3, @1, @3);
			}
		| const_seq DUTOK_IDENTIFIER DUTOK_EQ sign_add_opt DUTOK_UINT DUTOK_SEMICOLON	{
				mlc::make_const_int(ctx, $2, $4, $5, @2);
			}
		| const_seq DUTOK_IDENTIFIER DUTOK_EQ sign_add_opt DUTOK_REAL DUTOK_SEMICOLON {
				mlc::make_const_real(ctx, $2, $4, $5, @2);
			}
		| const_seq DUTOK_IDENTIFIER DUTOK_EQ DUTOK_STRING DUTOK_SEMICOLON	{
			  	ctx->tab->add_const_str(@2, $2, $4);
		  	}
		| const_seq DUTOK_IDENTIFIER DUTOK_EQ DUTOK_IDENTIFIER DUTOK_SEMICOLON	{
				mlc::make_const_from_id(ctx, $2, $4, @2, @4);
			}
		;

type_seq:	  DUTOK_IDENTIFIER DUTOK_EQ type_val DUTOK_SEMICOLON	{
				ctx->tab->add_type(@1, $1, $3);
			}
		| type_seq DUTOK_IDENTIFIER DUTOK_EQ type_val DUTOK_SEMICOLON	{
				ctx->tab->add_type(@2, $2, $4);
			}
		;

var_seq:	  identifier_seq DUTOK_COLON type_val DUTOK_SEMICOLON	{
				var_from_vec(ctx, $1, $3);
			}
		| identifier_seq DUTOK_COLON type_val DUTOK_SEMICOLON var_seq	{
				var_from_vec(ctx, $1, $3);
			}
		;

identifier_seq:	  DUTOK_IDENTIFIER	{
				$$.push_back(std::make_pair< mlc::ls_id_type::const_pointer, int>($1,@1));
			}

		| DUTOK_IDENTIFIER DUTOK_COMMA identifier_seq	{
				$$.push_back(std::make_pair< mlc::ls_id_type::const_pointer, int>($1,@1));
				$$.insert($$.end(), $3.begin(), $3.end());
			}
		;


/* type machinery */
type_val:	  ord_type_noid		{ $$ = $1; }
		| struc_type_val	{ $$ = $1;}
		| DUTOK_IDENTIFIER	{
				mlc::symbol_pointer sp = ctx->tab->find_symbol($1);
				type_pointer tp = sp->access_typed()->type();
				if (sp->kind() != SKIND_TYPE) { error(DUERR_NOTTYPE, @1, *$1); }
				$$ = tp;
			}
		| DUTOK_IDENTIFIER DUTOK_DOTDOT ord_const_id	{
				$$ = range_from_id_ord_const(ctx, $1, $3, @1);
			}
		;

ord_type_id:	  ord_const_noid DUTOK_DOTDOT ord_const_id {
				$$ = range_from_ord_const_ord_const(ctx, $1, $3, @1);
			}
		| DUTOK_IDENTIFIER DUTOK_DOTDOT ord_const_id	{
				$$ = range_from_id_ord_const(ctx, $1, $3, @1);
			}
		| DUTOK_IDENTIFIER	{
				mlc::symbol_pointer sp = ctx->tab->find_symbol($1);
				type_pointer tp = sp->access_typed()->type();
				if (sp->kind() != SKIND_TYPE) { error(DUERR_NOTTYPE, @1, *$1); }
				$$ = tp;
			}
		;

ord_type_noid:	  ord_const_noid DUTOK_DOTDOT ord_const_id	{
				$$ = range_from_ord_const_ord_const(ctx, $1, $3, @1);
			}
		;

struc_type_val:	  DUTOK_ARRAY DUTOK_LSBRA ord_type_seq DUTOK_RSBRA DUTOK_OF type_val	{
				mlc::type_pointer et; /* index type, element type */;
				et = $6;
				/* ord_type_seq size >= 1 */
				for( tp_vec_riter rit = $3.rbegin(); rit != $3.rend(); ++rit) {
					et = ctx->tab->create_array_type(*rit, et);
				}
				$$ = et;
			}
		;

ord_type_seq:	  ord_type_id	{
				$$.clear();
				$$.push_back($1);
			}
		| ord_type_seq DUTOK_COMMA ord_type_id {
				/* pass range sequence */
				/*for(tp_vec_iter it = $1.begin(); it != $1.end(); ++it) {
					$$.push_back(*it);
				}*/
				$$.push_back($3);				
			}
		;

ord_const_noid:	  DUTOK_OPER_SIGNADD DUTOK_IDENTIFIER	{ /* identifier of const integer */
				$$ = ord_const_from_id(ctx, $1, $2, @2);
			}
		| sign_add_opt DUTOK_UINT	{
				$$ = ord_const_from_uint(ctx, $1, $2);
			}
		;

ord_const_id:	  sign_add_opt DUTOK_IDENTIFIER	{ /* identifier of const integer */
				$$ = ord_const_from_id(ctx, $1, $2, @2);
			}
		| sign_add_opt DUTOK_UINT {
				$$ = ord_const_from_uint(ctx, $1, $2);
			}
		;

/* stmt machinery */

stmt:		  matched	{ $$ = $1; }
		| unmatched	{ $$ = $1; }
		;

matched:	  label_inst_opt DUTOK_IF expr DUTOK_THEN matched DUTOK_ELSE matched	{
				add_if(ctx, $1, $3, @3, $5, $7);
				$$ = $1;
			}
		| label_inst_opt variable /* array variable */ DUTOK_ASSIGN expr	{
				add_xstx(ctx, $1, $2, $4);
				$$ = $1;
			}
		| label_inst_opt DUTOK_IDENTIFIER /* identifier of function or variable */ DUTOK_ASSIGN expr	{
				mlc::symbol_pointer sp = ctx->tab->find_symbol($2);
				if(sp->kind() == mlc::SKIND_FUNCTION) {
					expr_deref(ctx, $4);
					if(ctx->tab->nested() && (ctx->tab->my_function_name() == $2)) {
						expr_deptr($4);
						add_fncret(ctx, $4, $2, @2);
					} else {
						error(DUERR_NOTLVALUE, @2);
					}
					mlc::icblock_append_delete($1, $4.icb_p_);
					$$ = $1;
				} else if (sp->kind() == SKIND_GLOBAL_VARIABLE) {
					/*if(sp->access_typed()->type()->cat() == TCAT_ARRAY) {
						assert(false);

						$$ = $1;
					} else {*/
						expr_deref(ctx, $4);
						//expr_deptr($4);
						add_gstx(ctx, $4, $2, @2);
						//$$ = $4.icb_p_;
						mlc::icblock_append_delete($1, $4.icb_p_);
						$$ = $1;
					//}
				} else if (sp->kind() == SKIND_LOCAL_VARIABLE) {
					expr_deref(ctx, $4);
					expr_deptr($4);
					add_lstx(ctx, $4, $2, @2);
					//$$ = $4.icb_p_;
					mlc::icblock_append_delete($1, $4.icb_p_);
					$$ = $1;
				} else if (sp->kind() == SKIND_PARAMETER_BY_REFERENCE) {
					expr_deref(ctx, $4);
					expr_deptr($4);
					add_xstx_param(ctx, $1, sp->access_parameter_by_reference(), $4);
					$$ = $1;
				} else {
					error(DUERR_NOTLVALUE, @2);
				}
			}
		| label_inst_opt DUTOK_IDENTIFIER /* identifier of procedure or function */	{
				mlc::symbol_pointer sp = ctx->tab->find_symbol($2);
				if(sp->kind() == mlc::SKIND_PROCEDURE) {
					expr_vec foo = expr_vec();
					call_proc(ctx, $1, sp->access_subprogram(), foo, $2, @2);
					$$ = $1;
				} else if(sp->kind() == mlc::SKIND_FUNCTION) {
					expr_vec foo;
					call_fun(ctx, $1, sp->access_subprogram(), foo, $2, @2, true);
					$$ = $1;
				} else {
					error(DUERR_NOTPROC, @2, *$2);
					$$ = $1;
				}
			}
		| label_inst_opt DUTOK_IDENTIFIER /* identifier of procedure or function */ DUTOK_LPAR real_params DUTOK_RPAR	{
				mlc::symbol_pointer sp = ctx->tab->find_symbol($2);
				if(sp->kind() == mlc::SKIND_PROCEDURE) {
					call_proc(ctx, $1, sp->access_subprogram(), $4, $2, @2);
					$$ = $1;
				} else if(sp->kind() == mlc::SKIND_FUNCTION) {
					call_fun(ctx, $1, sp->access_subprogram(), $4, $2, @2, true); 
					$$ = $1;
				} else {
					error(DUERR_NOTPROC, @2, *$2);
					$$ = $1;
				}
			}
		| label_inst_opt DUTOK_GOTO DUTOK_UINT	{
				mlc::label_symbol_pointer lp = ctx->tab->find_label($3); 
				if(!lp) {
					error(mlc::DUERR_NOTLABEL, @3, *$3);
				} else {
					$1->append_instruction_with_target( new ai::JMP($1->end()), lp->label());
					lp->goto_found(@3);
				}

				$$ = $1;
				
			}
		| label_inst_opt DUTOK_BEGIN stmt_seq DUTOK_END	{
				mlc::icblock_append_delete($1, $3);
				$$ = $1;
			}
		| label_inst_opt DUTOK_REPEAT matched_seq DUTOK_UNTIL expr /* expr of bool */	{
				add_repeat(ctx, $1, $3, $5, @5);
				$$ = $1;
			}
		| label_inst_opt DUTOK_FOR DUTOK_IDENTIFIER /* identifier of ordinal type variable */ DUTOK_ASSIGN expr /* ordinal type expr */ DUTOK_FOR_DIRECTION expr /* expr of ordinal */ DUTOK_DO matched	{
				add_for(ctx, $1, $3, @3, $5, $6, $7, $9);
				$$ = $1;
			}
		| label_inst_opt DUTOK_WHILE expr /* expr of bool */ DUTOK_DO matched	{
				add_while(ctx, $1, $3, @3, $5);
				$$ = $1;
			}
		| label_inst_opt /* void stmt */	{
				$$ = $1; //mlc::icblock_create();
			}
		;
		
unmatched:	  label_inst_opt DUTOK_IF expr /* expr of bool */ DUTOK_THEN stmt	{
				add_if(ctx, $1, $3, @3, $5, mlc::icblock_create());
				$$ = $1;
			}
		| label_inst_opt DUTOK_IF expr /* expr of bool */ DUTOK_THEN matched DUTOK_ELSE unmatched	{
				add_if(ctx, $1, $3, @3, $5, $7);
				$$ = $1;
			}
		| label_inst_opt DUTOK_FOR DUTOK_IDENTIFIER /* identifier of ordinal type variable */ DUTOK_ASSIGN expr /* ordinal type expr */ DUTOK_FOR_DIRECTION expr /* expr of ordinal */ DUTOK_DO unmatched /* ?? matched ?? */	{
				add_for(ctx, $1, $3, @3, $5, $6, $7, $9);
				$$ = $1;
			}
		| label_inst_opt DUTOK_WHILE expr /* expr of bool */ DUTOK_DO unmatched	{
				add_while(ctx, $1, $3, @3, $5);
				$$ = $1;
			}
		| label_inst_opt DUTOK_REPEAT unmatched_seq DUTOK_UNTIL expr /* expr of bool */	{
				add_repeat(ctx, $1, $3, $5, @5);
				$$ = $1;
			}
		;

stmt_seq:	  stmt	{ $$ = $1; }
		| stmt DUTOK_SEMICOLON stmt_seq	{
				mlc::icblock_append_delete($1, $3); //appends $3 to $1, kills $1
				$$ = $1;
			}
		;

matched_seq:	  matched	{
				$$ = $1;
			}	
		| matched DUTOK_SEMICOLON matched_seq	{
				mlc::icblock_append_delete($1, $3);
				$$ = $1;
			}
		;

unmatched_seq:	  unmatched	{
				$$ = $1;
			}
		| unmatched DUTOK_SEMICOLON unmatched_seq	{
				mlc::icblock_append_delete($1, $3);
				$$ = $1;
			}
		;

/* expr machinery */

expr:		  simple_expr	{
				$$ = $1;
			}
		| simple_expr DUTOK_OPER_REL simple_expr	{
				add_rel(ctx, $1, $2, @2, $3);
				$$ = $1;
			}
		| simple_expr DUTOK_EQ simple_expr {
				add_eq(ctx, $1, @2, $3);
				$$ = $1;
			}
		;

simple_expr:	  term	{
				$$ = $1;
			}
		| DUTOK_OPER_SIGNADD term	{
				/* TODO */
				if($1 == DUTOKGE_MINUS) {
					expr_deref(ctx, $2);
					expr_deptr($2);
					add_minus($2, @2);
				}
				$$ = $2;
			}
		| simple_expr DUTOK_OPER_SIGNADD term {
				if($2 == DUTOKGE_PLUS) {
					$$.init_unref();
					expr_add(ctx, $$, $1, $3, @2);
				} else if($2 == DUTOKGE_MINUS) {
					$$.init_unref();
					expr_sub(ctx, $$, $1, $3, @2);
				} else {
					assert(false);
				}

			}
		| simple_expr DUTOK_OR term {
				$$.init_unref();
				expr_or(ctx, $$, $1, $3, @2);
			}
		;


term:		  faktor	{
				$$ = $1;
			}
		| faktor DUTOK_OPER_MUL term	{
				if($2 == DUTOKGE_ASTERISK) {
					$$.init_unref();
					expr_mul(ctx, $$, $1, $3, @2);
				} else if ($2 == DUTOKGE_SOLIDUS) {
					$$.init_unref();
					expr_divr(ctx, $$, $1, $3, @2);
				} else if ($2 == DUTOKGE_DIV) {
					$$.init_unref();
					expr_divi(ctx, $$, $1, $3, @2);
				} else if ($2 == DUTOKGE_MOD) {
					$$.init_unref();
					expr_mod(ctx, $$, $1, $3, @2);
				} else if ($2 == DUTOKGE_AND) {
					$$.init_unref();
					expr_and(ctx, $$, $1, $3, @2);
				} else {
					assert(false);
				}

				
			}
		;

faktor:		  DUTOK_UINT	{
				$$.init_unref();
				$$.icb_p_->append_instruction( new ai::LDLITI($1));
				$$.extp_ = ctx->tab->logical_integer();
			}
		| DUTOK_REAL	{
				$$.init_unref();
				$$.icb_p_->append_instruction( new ai::LDLITR($1));
				$$.extp_ = ctx->tab->logical_real();
			}
		| DUTOK_STRING	{
				$$.init_unref();
				$$.icb_p_->append_instruction( new ai::LDLITS($1));
				$$.extp_ = ctx->tab->logical_string();
			}
		| variable	{
				/* TODO array variable */
				$$ = $1;
			}
		| DUTOK_IDENTIFIER /* identifier of variable or const (ie even bool) or function */	{
				$$.byref_ = true; //still possibly dereferenceable
				$$.id_ci_ = $1;
				$$.id_l_ = @1;
				
			}
		| DUTOK_IDENTIFIER /*identifier of function */ DUTOK_LPAR real_params DUTOK_RPAR	{
				mlc::symbol_pointer sp = ctx->tab->find_symbol($1);
				$$.init_unref();
				$$.extp_ = sp->access_typed()->type();

				if(sp->kind() == SKIND_FUNCTION) {
					call_fun(ctx, $$.icb_p_, sp->access_subprogram(), $3, $1, @1); 

				} else {
					error(DUERR_NOTPROC, @1, *$1);
				}
			}
		| DUTOK_LPAR expr DUTOK_RPAR	{
				$$ = $2;
			}
		| DUTOK_NOT faktor	{
				/* TODO logical not */
			}
		;

real_params:	expr	{
				$$.push_back($1);
			}
		| expr DUTOK_COMMA real_params	{
			$$ = $3;
			$$.push_back($1);

		}
		;

variable:	  DUTOK_IDENTIFIER /* identifier of array variable */ DUTOK_LSBRA ordinal_expr_seq DUTOK_RSBRA	{
				/* TODO */
				$$.is_p_ = true;
				$$.id_ci_ = $1;
				$$.id_l_ = @1;
				$$.icb_p_ = mlc::icblock_create();

				mlc::symbol_pointer sp = ctx->tab->find_symbol($1);
				$$.extp_ = sp->access_typed()->type()->access_array();
				
				if(sp->access_typed()->type()->cat() == TCAT_ARRAY) {
					mlc::array_type_pointer arr = sp->access_typed()->type()->access_array();
					arr_index(ctx, $$, $3, false);
				} else {
					mlc::error(DUERR_NOTARRAY, @1, *$1);
				}

			}
		| variable /* identifier of array variable */ DUTOK_LSBRA ordinal_expr_seq DUTOK_RSBRA	{
				/* TODO */
				$$ = $1;

				arr_index(ctx, $$, $3, true);
			}
		;

ordinal_expr_seq: expr /* expr of ordinal type */	{
				$$.push_back($1);
			}
		| expr DUTOK_COMMA ordinal_expr_seq	{
				$3.push_back($1);
				$$ = $3;
			}
		;



sign_add_opt:	  DUTOK_OPER_SIGNADD	{ $$ = $1;}
		| /* no sign add */	{ $$ = DUTOKGE_PLUS;}
		;

label_inst_opt:	  DUTOK_UINT DUTOK_COLON	{ 
				$$ = mlc::icblock_create();
				mlc::label_symbol_pointer lp = ctx->tab->find_label($1); 
				if(!lp) {
					error(mlc::DUERR_NOTLABEL, @1, *$1);
				} else {
					$$->add_label(lp->label());
					lp->label_found(@1);
				}
			}
		| /* no label */	{
				$$ = mlc::icblock_create();
			}
		;

%%

static void yyerror(unsigned *line, MlaskalCtx *, const char *m)
{
    error(DUERR_SYNTAX, *line, m);
}
