/*
 *	frantisek farka
 */

#include "du6sem.h"
#include "du6tok.h"
#include "du6.h"
#include "duerr.h"

namespace mlc {


	void 
	make_const_from_id(MlaskalCtx* ctx, mlc::ls_id_type::const_pointer id, mlc::ls_id_type::const_pointer val, int l_id, int l_val) {

		mlc::symbol_pointer sp = ctx->tab->find_symbol(val);
		if ( sp->kind() != SKIND_CONST ) {
			error( DUERR_NOTCONST, l_val, *val);
		} else {
			bool bval;
			mlc::ls_int_index iptr;
			mlc::ls_real_index rptr;
			mlc::ls_str_index sptr;
			switch(sp->access_typed()->type()->cat()) {
				case mlc::TCAT_BOOL:
					bval = sp->access_const()->access_bool_const()->bool_value();
					ctx->tab->add_const_bool(l_id, id, bval);
					break;
				case mlc::TCAT_INT:
					iptr = sp->access_const()->access_int_const()->int_value();
					ctx->tab->add_const_int(l_id, id, iptr);
					break;
				case mlc::TCAT_REAL:
					rptr = sp->access_const()->access_real_const()->real_value();
					ctx->tab->add_const_real(l_id, id, rptr);
					break;
				case mlc::TCAT_STR:
					sptr = sp->access_const()->access_str_const()->str_value();
					ctx->tab->add_const_str(l_id, id, sptr);
					break;
				case TCAT_UNDEF:
				case TCAT_ARRAY:
				case TCAT_RECORD:
				case TCAT_RANGE:
				default:
					assert(false);
					break;
			}
		}

	}

	void
	make_const_int(mlc::MlaskalCtx* ctx, mlc::ls_id_type::const_pointer id, int sign, mlc::ls_int_type::const_pointer val, int l_id) {

		switch(sign) {
			case DUTOKGE_PLUS:
				ctx->tab->add_const_int(l_id, id, val);
				break;
			case DUTOKGE_MINUS:
				mlc::ls_int_type::const_pointer nval = ctx->tab->ls_int().add( - *val);
				ctx->tab->add_const_int(l_id, id, nval);
				break;
		}
	}

	void
	make_const_real(mlc::MlaskalCtx* ctx, mlc::ls_id_type::const_pointer id, int sign, mlc::ls_real_type::const_pointer val, int l_id) {
		switch(sign) {
			case DUTOKGE_PLUS:
				ctx->tab->add_const_real(l_id, id, val);
				break;
			case DUTOKGE_MINUS:
				mlc::ls_real_type::const_pointer nval = ctx->tab->ls_real().add( - *val);
				ctx->tab->add_const_real(l_id, id, nval);
				break;
		}
	}


	mlc::ls_int_type::const_pointer
	ord_const_from_id(MlaskalCtx* ctx, int sign, mlc::ls_id_type::const_pointer id, int l_id) {

		mlc::symbol_pointer sp = ctx->tab->find_symbol(id);
		//mlc::ls_int_type::const_pointer ci;

		if ( sp->kind() != SKIND_CONST ) {
			error( DUERR_NOTCONST, l_id, *id);
		} else if(sp->access_typed()->type()->cat() != TCAT_INT) {
			error( DUERR_NOTINTEGRAL, l_id, *id);
			//TODO character const?
		} else {
			switch(sign) {
				case DUTOKGE_PLUS:
					return sp->access_const()->access_int_const()->int_value();
					break;
				case DUTOKGE_MINUS:
					return ctx->tab->ls_int().add( - *sp->access_const()->access_int_const()->int_value());
					break;
			}
		}
		return sp->access_const()->access_int_const()->int_value(); // just to shut them up
	}

	mlc::ls_int_type::const_pointer
	ord_const_from_uint(MlaskalCtx* ctx, int sign, mlc::ls_int_type::const_pointer int_ci) {
		if(sign == DUTOKGE_MINUS) {
			return ctx->tab->ls_int().add( - *int_ci);
		}

		return int_ci;
	}

	mlc::type_pointer
	range_from_id_ord_const(mlc::MlaskalCtx* ctx, mlc::ls_id_type::const_pointer id_lower, mlc::ls_int_type::const_pointer int_upper, int l_lower )  {

		mlc::ls_int_type::const_pointer int_lower = ord_const_from_id(ctx, DUTOKGE_PLUS, id_lower, l_lower);

		if(*int_lower > *int_upper) {
			error( DUERR_BADRANGE, l_lower);
		}

		// create range, even if it is invalid
		return ctx->tab->create_range_type(int_lower, int_upper);
	}

	mlc::type_pointer
	range_from_ord_const_ord_const(mlc::MlaskalCtx* ctx, mlc::ls_int_type::const_pointer int_lower, mlc::ls_int_type::const_pointer int_upper, int l_lower)  {

		if(*int_lower > *int_upper) {
			error( DUERR_BADRANGE, l_lower);
		}

		// create range, even if it is invalid
		return ctx->tab->create_range_type(int_lower, int_upper);
	}

	void
	var_from_vec(mlc::MlaskalCtx* ctx, const id_ci_vec & vars, mlc::type_pointer type) {

		for(id_ci_vec_criter rit = vars.rbegin(); rit != vars.rend(); ++rit) {
			ctx->tab->add_var(rit->second, rit->first, type);
		}
	}

	mlc::type_pointer
	get_scalar_type(MlaskalCtx* ctx, mlc::ls_id_type::const_pointer id, int l_id) {
		mlc::symbol_pointer sp = ctx->tab->find_symbol(id);
		if(sp->kind() != mlc::SKIND_TYPE) {
			error(DUERR_NOTTYPE, l_id, *id);
			return	ctx->tab->logical_undef();
		}
		switch(sp->access_typed()->type()->cat()) {
			case TCAT_BOOL:
			case TCAT_INT:
			case TCAT_REAL:
			case TCAT_STR:
			case TCAT_RANGE:
				return sp->access_type()->type();
				break;
			case TCAT_UNDEF:
			case TCAT_ARRAY:
			case TCAT_RECORD:
				error(DUERR_NOTSCALAR, l_id, *id);
				return	ctx->tab->logical_undef();
				break;
		}

		return	ctx->tab->logical_undef();
	}

	mlc::type_pointer
	get_type(MlaskalCtx* ctx, mlc::ls_id_type::const_pointer id, int l_id) {
		mlc::symbol_pointer sp = ctx->tab->find_symbol(id);
		if(sp->kind() != SKIND_TYPE) {
			error(DUERR_NOTTYPE, l_id, *id);
		}

		return sp->access_typed()->type();
	}

	void
	params_append_byref(mlc::parameter_list* params, const id_ci_vec & ids, mlc::type_pointer type) {
		for(id_ci_vec_citer it = ids.begin(); it != ids.end(); ++it) {
			params->append_parameter_by_reference(it->first, type);
		}
	}

	void
	params_append_byval(mlc::parameter_list* params, const id_ci_vec & ids, mlc::type_pointer type) {
		for(id_ci_vec_citer it = ids.begin(); it != ids.end(); ++it) {
			params->append_parameter_by_value(it->first, type);
		}
	}

	/************************************************************************************************/
	void array_deref(MlaskalExpr & expr, mlc::symbol_pointer sp) {
		mlc::icblock_pointer ip = mlc::icblock_create();

		if(sp->kind() == SKIND_GLOBAL_VARIABLE) {
			ip->append_instruction( new ai::GREF(sp->access_variable()->address()));
		} else if(sp->kind() == SKIND_LOCAL_VARIABLE) {
			ip->append_instruction( new ai::LREF(sp->access_variable()->address()));
		} else {
			assert(false);
		}

		if(expr.icb_p_ == NULL) {
			expr.icb_p_ = ip;
			expr.extp_ = sp->access_typed()->type();
		} else {
			//hnusny hack
			if(expr.icb_p_->size() == 1) {
				delete expr.icb_p_;
				expr.icb_p_ = ip;
			} else  {
				mlc::icblock_append_delete(ip, expr.icb_p_);
				expr.icb_p_ = ip;
				expr.icb_p_->append_instruction( new ai::ADDP());
			}

		}

	}

	void expr_deptr(MlaskalExpr & expr) {
		if(expr.is_p_) {
			add_xldx(expr);
			expr.is_p_ = false;
		}
	}

	void expr_deref(MlaskalCtx* ctx, MlaskalExpr & expr) {
		if(expr.byref_) {

			mlc::symbol_pointer sp = ctx->tab->find_symbol(expr.id_ci_);

			if(sp->kind() == SKIND_FUNCTION) {
					expr.init_unref();
					expr_vec foo;
					call_fun(ctx, expr.icb_p_, sp->access_subprogram(), foo, expr.id_ci_, expr.id_l_);
					expr.extp_ = sp->access_typed()->type();

			} else if (sp->kind() == SKIND_GLOBAL_VARIABLE) {
				mlc::variable_symbol_pointer gv = ctx->tab->find_symbol(expr.id_ci_)->access_variable();
				if(sp->access_typed()->type()->cat() == TCAT_INT) {
					expr.init_unref();
					expr.icb_p_->append_instruction( new ai::GLDI(gv->address()));
					expr.extp_ = sp->access_typed()->type();

				} else if(sp->access_typed()->type()->cat() == TCAT_REAL) {
					expr.init_unref();
					expr.icb_p_->append_instruction( new ai::GLDR(gv->address()));
					expr.extp_ = sp->access_typed()->type();

				} else if(sp->access_typed()->type()->cat() == TCAT_BOOL) {
					expr.init_unref();
					expr.icb_p_->append_instruction( new ai::GLDB(gv->address()));
					expr.extp_ = sp->access_typed()->type();

				} else if(sp->access_typed()->type()->cat() == TCAT_STR) {
					expr.init_unref();
					expr.icb_p_->append_instruction( new ai::GLDS(gv->address()));
					expr.extp_ = sp->access_typed()->type();

				} else if(sp->access_typed()->type()->cat() == TCAT_ARRAY) {
					array_deref(expr, sp);

				} else {
					/* TODO could be other types */
					error(DUERR_CANNOTCONVERT, expr.id_l_);
					expr.init_unref();
					expr.extp_ = ctx->tab->logical_undef();
				}
				
			} else if (sp->kind() == SKIND_LOCAL_VARIABLE) {
				mlc::variable_symbol_pointer lv = ctx->tab->find_symbol(expr.id_ci_)->access_variable();
				if(sp->access_typed()->type()->cat() == TCAT_INT) {
					expr.init_unref();
					expr.icb_p_->append_instruction( new ai::LLDI(lv->address()));
					expr.extp_ = sp->access_typed()->type();

				} else if(sp->access_typed()->type()->cat() == TCAT_REAL) {
					expr.init_unref();
					expr.icb_p_->append_instruction( new ai::LLDR(lv->address()));
					expr.extp_ = sp->access_typed()->type();

				} else if(sp->access_typed()->type()->cat() == TCAT_BOOL) {
					expr.init_unref();
					expr.icb_p_->append_instruction( new ai::LLDB(lv->address()));
					expr.extp_ = sp->access_typed()->type();

				} else if(sp->access_typed()->type()->cat() == TCAT_STR) {
					expr.init_unref();
					expr.icb_p_->append_instruction( new ai::LLDS(lv->address()));
					expr.extp_ = sp->access_typed()->type();

				} else if(sp->access_typed()->type()->cat() == TCAT_ARRAY) {
					array_deref(expr, sp);
				} else {
					/* TODO could be other types */
					error(DUERR_CANNOTCONVERT, expr.id_l_); 
					expr.init_unref();
					expr.extp_ = ctx->tab->logical_undef();
				}

			} else if (sp->kind() == SKIND_CONST) {
				if(sp->access_typed()->type()->cat() == TCAT_INT) {
					expr.init_unref();
					mlc::int_const_symbol_pointer ic = ctx->tab->find_symbol(expr.id_ci_)->access_const()->access_int_const();
					expr.icb_p_->append_instruction( new ai::LDLITI(ic->int_value()));
					expr.extp_ = sp->access_typed()->type();

				} else if(sp->access_typed()->type()->cat() == TCAT_REAL) {
					expr.init_unref();
					mlc::real_const_symbol_pointer rc = ctx->tab->find_symbol(expr.id_ci_)->access_const()->access_real_const();
					expr.icb_p_->append_instruction( new ai::LDLITR(rc->real_value()));
					expr.extp_ = sp->access_typed()->type();

				} else if(sp->access_typed()->type()->cat() == TCAT_BOOL) {
					expr.init_unref();
					mlc::bool_const_symbol_pointer bc = ctx->tab->find_symbol(expr.id_ci_)->access_const()->access_bool_const();
					expr.icb_p_->append_instruction( new ai::LDLITB(bc->bool_value()));
					expr.extp_ = sp->access_typed()->type();

				} else if(sp->access_typed()->type()->cat() == TCAT_STR) {
					expr.init_unref();
					mlc::str_const_symbol_pointer sc = ctx->tab->find_symbol(expr.id_ci_)->access_const()->access_str_const();
					expr.icb_p_->append_instruction( new ai::LDLITS(sc->str_value()));
					expr.extp_ = sp->access_typed()->type();

				} else {
					error(DUERR_CANNOTCONVERT, expr.id_l_);
					expr.init_unref();
					expr.extp_ = ctx->tab->logical_undef();
				}
			} else if (sp->kind() == SKIND_PARAMETER_BY_REFERENCE) {

				//mlc::parameter_by_reference_symbol_pointer pbref = sp->access_parameter_by_reference();
				expr.init_unref();
				expr.icb_p_->append_instruction( new ai::LLDP(sp->access_parameter_by_reference()->address()));
				expr.extp_ = sp->access_parameter_by_reference()->access_typed()->type();

				if(sp->access_parameter_by_reference()->access_typed()->type()->cat() == TCAT_INT) {
					expr.icb_p_->append_instruction( new ai::XLDI());
				} else if(sp->access_parameter_by_reference()->access_typed()->type()->cat() == TCAT_REAL) {
					expr.icb_p_->append_instruction( new ai::XLDR());
				} else if(sp->access_parameter_by_reference()->access_typed()->type()->cat() == TCAT_STR) {
					expr.icb_p_->append_instruction( new ai::XLDS());
				} else if(sp->access_parameter_by_reference()->access_typed()->type()->cat() == TCAT_INT) {
					expr.icb_p_->append_instruction( new ai::XLDB());
				} else if(sp->access_parameter_by_reference()->access_typed()->type()->cat() == TCAT_ARRAY) {

				} else {
					assert(false && "parameters by reference types");
				}

			} else {
				error(DUERR_NOTVAR, expr.id_l_, *(expr.id_ci_));
				expr.init_unref();
				expr.extp_ = ctx->tab->logical_undef();
			}

		}

		//expr_deptr(expr);
	}

	void
	add_gstx(MlaskalCtx* ctx, MlaskalExpr & expr, mlc::ls_id_type::const_pointer id, int l_id) {

		mlc::global_variable_symbol_pointer gv = ctx->tab->find_symbol(id)->access_global_variable();

		switch(expr.extp_->cat()) {
			case TCAT_BOOL:
				if(gv->access_typed()->type()->cat() == TCAT_BOOL) {
					expr.icb_p_->append_instruction(new ai::GSTB(gv->address()));
				} else {
					error(DUERR_CANNOTCONVERT, l_id); 
				}
				break;
			case TCAT_INT:
				/* expression is of integer type, store to: */
				if(gv->access_typed()->type()->cat() == TCAT_INT) {
					expr.icb_p_->append_instruction(new ai::GSTI(gv->address()));
				} else if(gv->access_typed()->type()->cat() == TCAT_REAL) {
					/* silent convert */
					expr.icb_p_->append_instruction(new ai::CVRTIR());
					expr.icb_p_->append_instruction(new ai::GSTR(gv->address()));
				} else {
					error(DUERR_CANNOTCONVERT, l_id); 
				}
				break;
			case TCAT_REAL:
				/* expression is of real type, store to: */
				if(gv->access_typed()->type()->cat() == TCAT_REAL) {
					expr.icb_p_->append_instruction(new ai::GSTR(gv->address()));
				} else if(gv->access_typed()->type()->cat() == TCAT_INT) {
					/* verbose convert */
					expr.icb_p_->append_instruction(new ai::CVRTRI());
					expr.icb_p_->append_instruction(new ai::GSTI(gv->address()));
					error(DUERR_CONVERSION, l_id);
				} else {
					error(DUERR_CANNOTCONVERT, l_id); 
				}
				break;
			case TCAT_STR:
				if(gv->access_typed()->type()->cat() == TCAT_STR) {
					expr.icb_p_->append_instruction(new ai::GSTS(gv->address()));
				} else {
					error(DUERR_CANNOTCONVERT, l_id); 
				}
				break;
			case TCAT_RANGE:
					/* forever undef */
					assert(false);
				break;
			case TCAT_ARRAY:
					if(mlc::identical_type(gv->access_typed()->type(), expr.extp_)) {
						//on stack pointer to array
						//save as local variable
						expr.icb_p_->append_instruction( new ai::SSTP(10));

						for(int i = 0; i < expr.extp_->compute_size(); ++i) {
							expr.icb_p_->append_instruction( new ai::SLDP(11));
							//TODO
							if(expr.extp_->compute_final_ptype() == mlaskal::PTYPE_INTEGER) {
								expr.icb_p_->append_instruction( new ai::XLDI());
							} else {
								assert(false);
							}
							expr.icb_p_->append_instruction( new ai::GREF(gv->address() + i));
							if(expr.extp_->compute_final_ptype() == mlaskal::PTYPE_INTEGER) {
								expr.icb_p_->append_instruction( new ai::XSTI());
							} else {
								assert(false);
							}

							//shift pointer
							expr.icb_p_->append_instruction( new ai::SLDP(11));
							expr.icb_p_->append_instruction( new ai::LDLITI(ctx->tab->one()));
							expr.icb_p_->append_instruction( new ai::ADDP());
							expr.icb_p_->append_instruction( new ai::SSTP(10));

						}
						//expr.icb_p_->append_instruction( new ai::SLDP(11));
						//expr.icb_p_->append_instruction( new ai::SLDP(10));

					} else {
						error(DUERR_CANNOTCONVERT, l_id);
					}
				break;
			case TCAT_RECORD:
					/* forever undef */
					assert(false);
				break;
			case TCAT_UNDEF:
					/* wtf? */
					assert(false);
				break;
		}
	}

	void
	add_lstx(MlaskalCtx* ctx, MlaskalExpr & expr, mlc::ls_id_type::const_pointer id, int l_id) {

		mlc::local_variable_symbol_pointer lv = ctx->tab->find_symbol(id)->access_local_variable();

		switch(expr.extp_->cat()) {
			case TCAT_BOOL:
				if(lv->access_typed()->type()->cat() == TCAT_BOOL) {
					expr.icb_p_->append_instruction(new ai::LSTB(lv->address()));
				} else {
					error(DUERR_CANNOTCONVERT, l_id); 
				}
				break;
			case TCAT_INT:
				/* expression is of integer type, store to: */
				if(lv->access_typed()->type()->cat() == TCAT_INT) {
					expr.icb_p_->append_instruction(new ai::LSTI(lv->address()));
				} else if(lv->access_typed()->type()->cat() == TCAT_REAL) {
					/* silent convert */
					expr.icb_p_->append_instruction(new ai::CVRTIR());
					expr.icb_p_->append_instruction(new ai::LSTR(lv->address()));
				} else {
					error(DUERR_CANNOTCONVERT, l_id); 
				}
				break;
			case TCAT_REAL:
				/* expression is of real type, store to: */
				if(lv->access_typed()->type()->cat() == TCAT_REAL) {
					expr.icb_p_->append_instruction(new ai::LSTR(lv->address()));
				} else if(lv->access_typed()->type()->cat() == TCAT_INT) {
					/* verbose convert */
					expr.icb_p_->append_instruction(new ai::CVRTRI());
					expr.icb_p_->append_instruction(new ai::LSTI(lv->address()));
					error(DUERR_CONVERSION, l_id);
				} else {
					error(DUERR_CANNOTCONVERT, l_id); 
				}
				break;
			case TCAT_STR:
				if(lv->access_typed()->type()->cat() == TCAT_STR) {
					expr.icb_p_->append_instruction(new ai::LSTS(lv->address()));
				} else {
					error(DUERR_CANNOTCONVERT, l_id); 
				}
				break;
			case TCAT_RANGE:
					/* sofar undef */
					assert(false);
				break;
			case TCAT_ARRAY:
					if(mlc::identical_type(lv->access_typed()->type(), expr.extp_)) {
						//on stack pointer to array
						//save as local variable
						expr.icb_p_->append_instruction( new ai::SSTP(10));
						for(int i = 0; i < expr.extp_->compute_size(); ++i) {
							expr.icb_p_->append_instruction( new ai::SLDP(11));
							//TODO
							if(expr.extp_->compute_final_ptype() == mlaskal::PTYPE_INTEGER) {
								expr.icb_p_->append_instruction( new ai::XLDI());
							} else {
								assert(false);
							}
							expr.icb_p_->append_instruction( new ai::LREF(lv->address() + i));
							if(expr.extp_->compute_final_ptype() == mlaskal::PTYPE_INTEGER) {
								expr.icb_p_->append_instruction( new ai::XSTI());
							} else {
								assert(false);
							}

							//shift pointer
							expr.icb_p_->append_instruction( new ai::SLDP(11));
							expr.icb_p_->append_instruction( new ai::LDLITI(ctx->tab->one()));
							expr.icb_p_->append_instruction( new ai::ADDP());
							expr.icb_p_->append_instruction( new ai::SSTP(10));

						}

					} else {
						error(DUERR_CANNOTCONVERT, l_id);
					}
				break;
			case TCAT_RECORD:
					/* forever undef */
					assert(false);
				break;
			case TCAT_UNDEF:
					/* wtf? */
					assert(false);
				break;
		}

	}


	void
	add_fncret(MlaskalCtx* ctx, MlaskalExpr & expr, mlc::ls_id_type::const_pointer id, int l_id) {

		mlc::function_symbol_pointer fn = ctx->tab->find_symbol(id)->access_function();

		switch(expr.extp_->cat()) {
			case TCAT_BOOL:
				if(fn->access_typed()->type()->cat() == TCAT_BOOL) {
					expr.icb_p_->append_instruction(new ai::LSTB(ctx->tab->my_return_address()));
				} else {
					error(DUERR_CANNOTCONVERT, l_id); 
				}
				break;
			case TCAT_INT:
				/* expression is of integer type, store to: */
				if(fn->access_typed()->type()->cat() == TCAT_INT) {
					expr.icb_p_->append_instruction(new ai::LSTI(ctx->tab->my_return_address()));
				} else if(fn->access_typed()->type()->cat() == TCAT_REAL) {
					/* silent convert */
					expr.icb_p_->append_instruction(new ai::CVRTIR());
					expr.icb_p_->append_instruction(new ai::LSTR(ctx->tab->my_return_address()));
				} else {
					error(DUERR_CANNOTCONVERT, l_id); 
				}
				break;
			case TCAT_REAL:
				/* expression is of real type, store to: */
				if(fn->access_typed()->type()->cat() == TCAT_REAL) {
					expr.icb_p_->append_instruction(new ai::LSTR(ctx->tab->my_return_address()));
				} else if(fn->access_typed()->type()->cat() == TCAT_INT) {
					/* verbose convert */
					expr.icb_p_->append_instruction(new ai::CVRTRI());
					expr.icb_p_->append_instruction(new ai::LSTI(ctx->tab->my_return_address()));
					error(DUERR_CONVERSION, l_id);
				} else {
					error(DUERR_CANNOTCONVERT, l_id); 
				}
				break;
			case TCAT_STR:
				if(fn->access_typed()->type()->cat() == TCAT_STR) {
					expr.icb_p_->append_instruction(new ai::LSTS(ctx->tab->my_return_address()));
				} else {
					error(DUERR_CANNOTCONVERT, l_id); 
				}
				break;
			case TCAT_RANGE:
					/* sofar undef */
					assert(false);
				break;
			case TCAT_ARRAY:
					/* sofar undef */
					assert(false);
				break;
			case TCAT_RECORD:
					/* forever undef */
					assert(false);
				break;
			case TCAT_UNDEF:
					/* wtf? */
					assert(false);
				break;
		}

	}

	/* TODO by reference */
	void
	call_proc(MlaskalCtx* ctx, mlc::icblock_pointer icb_p, mlc::subprogram_symbol_pointer proc, expr_vec rpar, mlc::ls_id_type::const_pointer id, int l_id) {
		
		expr_vec::reverse_iterator rit = rpar.rbegin();
		mlc::parameter_list::const_iterator cit = proc->parameters()->begin();

		mlc::icblock_pointer foo;
		mlc::icblock_pointer dtrblc = mlc::icblock_create();

		for( ; (cit != proc->parameters()->end()) && (rit != rpar.rend()); ++cit, ++rit) {

			bool dtorp = false;


			if(cit->partype == mlc::PMODE_BY_REFERENCE) {

				if(rit->byref_) {
					mlc::symbol_pointer sp = ctx->tab->find_symbol(rit->id_ci_);
					if(mlc::identical_type(cit->ltype, sp->access_typed()->type())) {
						if(sp->kind() == SKIND_LOCAL_VARIABLE) {
							icb_p->append_instruction( new ai::LREF(sp->access_variable()->address()));
						} else if(sp->kind() == SKIND_GLOBAL_VARIABLE) {
							icb_p->append_instruction( new ai::GREF(sp->access_variable()->address()));
						} else if(sp->kind() == SKIND_PARAMETER_BY_REFERENCE) {
							icb_p->append_instruction( new ai::LLDP(sp->access_variable()->address()));
							//assert(false && "parameter by reference");
						} else {
							assert(false);
						}

						dtorp = true; //destruct pointer from stack
					} else {
						error(DUERR_TYPEMISMATCH, l_id);
					}
				} else {
					error(DUERR_NOTPARAMVAR, l_id);
				}
			} else {
			
				expr_deref(ctx, *rit);
				expr_deptr(*rit);
	
				mlc::icblock_append_delete(icb_p, rit->icb_p_);
				if(!mlc::identical_type(cit->ltype, rit->extp_)) {
	
					if((cit->ltype->cat() == mlc::TCAT_REAL) && rit->extp_->cat() == mlc::TCAT_INT) {
						/* silent convert */
						icb_p->append_instruction( new ai::CVRTIR());

					} else if((cit->ltype->cat() == mlc::TCAT_INT) && (rit->extp_->cat() == mlc::TCAT_REAL)) {
						/* verbose convert */
						icb_p->append_instruction( new ai::CVRTRI());
						error(DUERR_CONVERSION, l_id);
					} else {
						error(DUERR_CANNOTCONVERT, l_id);
					}

				}

				if(cit->ltype->cat() == TCAT_ARRAY) {
					//copy array on stack
					int stk = cit->ltype->compute_size();
					icb_p->append_instruction( new ai::SSTP(stk));
					stk++; //stack decreased

					for(int i = 0; i < cit->ltype->compute_size(); ++i) {
						icb_p->append_instruction( new ai::SLDP(stk--));
						
						if(cit->ltype->compute_final_ptype() == mlaskal::PTYPE_INTEGER) {
							icb_p->append_instruction( new ai::XLDI());
						} else {
							//TODO
							assert(false);
						}

						//shift pointer
						icb_p->append_instruction( new ai::SLDP(stk));
						icb_p->append_instruction( new ai::LDLITI(ctx->tab->one()));
						icb_p->append_instruction( new ai::ADDP());
						icb_p->append_instruction( new ai::SSTP(stk-1));

					}
				}

			}

			if (dtorp == true) {
				foo = mlc::icblock_create();
				foo->append_instruction( new ai::DTORP());
				mlc::icblock_append_delete(foo, dtrblc);
				dtrblc = foo;
			} else if(cit->ltype->cat() == mlc::TCAT_INT) {
				foo = mlc::icblock_create();
				foo->append_instruction( new ai::DTORI());
				mlc::icblock_append_delete(foo, dtrblc);
				dtrblc = foo;
			} else if(cit->ltype->cat() == mlc::TCAT_REAL) {
				foo = mlc::icblock_create();
				foo->append_instruction( new ai::DTORR());
				mlc::icblock_append_delete(foo, dtrblc);
				dtrblc = foo;
			} else if(cit->ltype->cat() == mlc::TCAT_BOOL) {
				foo = mlc::icblock_create();
				foo->append_instruction( new ai::DTORB());
				mlc::icblock_append_delete(foo, dtrblc);
				dtrblc = foo;
			} else if(cit->ltype->cat() == mlc::TCAT_STR) {
				foo = mlc::icblock_create();
				foo->append_instruction( new ai::DTORS());
				mlc::icblock_append_delete(foo, dtrblc);
				dtrblc = foo;
			} else if(cit->ltype->cat() == mlc::TCAT_ARRAY) {
				foo = mlc::icblock_create();
				for(int i = 0; i < cit->ltype->compute_size(); ++i) {
					if(cit->ltype->compute_final_ptype() == mlaskal::PTYPE_INTEGER) {
						foo->append_instruction( new ai::DTORI());
					} else {
						//TODO
						assert(false);
					}
				}
				mlc::icblock_append_delete(foo, dtrblc);
				dtrblc = foo;
			} else {
				assert(false);
			}

		}
	
		if((cit != proc->parameters()->end()) || (rit != rpar.rend())) {
			error(DUERR_PARNUM, l_id, *id);
		}
	
		icb_p->append_instruction( new ai::CALL(proc->code()));
		mlc::icblock_append_delete(icb_p, dtrblc);

	}

	void
	call_fun(MlaskalCtx* ctx, mlc::icblock_pointer icb_p, mlc::subprogram_symbol_pointer fun, expr_vec rpar, mlc::ls_id_type::const_pointer id, int l_id, bool destr_ret) {
	
		mlc::icblock_pointer dtrret = mlc::icblock_create();

		if(fun->access_function()->type()->cat() == TCAT_INT) {
			icb_p->append_instruction( new ai::INITI());
			if(destr_ret)
				dtrret->append_instruction( new ai::DTORI());

		} else if(fun->access_function()->type()->cat() == TCAT_REAL) {
			icb_p->append_instruction( new ai::INITB());
			if(destr_ret)
				dtrret->append_instruction( new ai::DTORR());

		} else if(fun->access_function()->type()->cat() == TCAT_BOOL) {
			icb_p->append_instruction( new ai::INITR());
			if(destr_ret)
				dtrret->append_instruction( new ai::DTORB());

		} else if(fun->access_function()->type()->cat() == TCAT_STR) {
			icb_p->append_instruction( new ai::INITS());
			if(destr_ret)
				dtrret->append_instruction( new ai::DTORS());
		} else if(fun->access_function()->type()->cat() == TCAT_RANGE) {
			assert(false);
			/* TODO */
		} else {
			assert(false);
		}

		/* tiny fake */
		call_proc(ctx, icb_p, fun, rpar, id, l_id);

		mlc::icblock_append_delete(icb_p, dtrret);	
	}

	void
	expr_add( MlaskalCtx* ctx, MlaskalExpr & res, MlaskalExpr & lop, MlaskalExpr & rop, int l_op) {
		expr_deref(ctx, lop);
		expr_deref(ctx, rop);
		expr_deptr(lop);
		expr_deptr(rop);
		res.is_p_ = false;

		if(lop.extp_->cat() == TCAT_INT) {
			if(rop.extp_->cat() == TCAT_INT) {
				mlc::icblock_append_delete(res.icb_p_, lop.icb_p_);
				mlc::icblock_append_delete(res.icb_p_, rop.icb_p_);
				res.icb_p_->append_instruction( new ai::ADDI());
				res.extp_ = lop.extp_;

			} else if(rop.extp_->cat() == TCAT_REAL) {
				/* silent convert to real */
				mlc::icblock_append_delete(res.icb_p_, lop.icb_p_);
				res.icb_p_->append_instruction( new ai::CVRTIR);
				mlc::icblock_append_delete(res.icb_p_, rop.icb_p_);
				res.icb_p_->append_instruction( new ai::ADDR());
				res.extp_ = rop.extp_;
				
			} else {
				error(DUERR_CANNOTCONVERT, l_op);
			}
		} else if(lop.extp_->cat() == TCAT_REAL) {
			if(rop.extp_->cat() == TCAT_INT) {
				mlc::icblock_append_delete(res.icb_p_, lop.icb_p_);
				/* silent convert to real */
				mlc::icblock_append_delete(res.icb_p_, rop.icb_p_);
				res.icb_p_->append_instruction( new ai::CVRTIR);
				res.icb_p_->append_instruction( new ai::ADDR());
				res.extp_ = lop.extp_;

			} else if(rop.extp_->cat() == TCAT_REAL) {
				mlc::icblock_append_delete(res.icb_p_, lop.icb_p_);
				mlc::icblock_append_delete(res.icb_p_, rop.icb_p_);
				res.icb_p_->append_instruction( new ai::ADDR());
				res.extp_ = rop.extp_;
				
			} else {
				error(DUERR_CANNOTCONVERT, l_op);
			}
		} else if(lop.extp_->cat() == TCAT_STR) {
			if(rop.extp_->cat() == TCAT_STR) {
				mlc::icblock_append_delete(res.icb_p_, lop.icb_p_);
				mlc::icblock_append_delete(res.icb_p_, rop.icb_p_);
				res.icb_p_->append_instruction( new ai::ADDS());
				res.extp_ = lop.extp_;

			} else {
				error(DUERR_CANNOTCONVERT, l_op);
			}
		} else {
			error(DUERR_CANNOTCONVERT, l_op);

		}
	}

	void
	expr_mul( MlaskalCtx* ctx, MlaskalExpr & res, MlaskalExpr & lop, MlaskalExpr & rop, int l_op) {
		expr_deref(ctx, lop);
		expr_deref(ctx, rop);
		expr_deptr(lop);
		expr_deptr(rop);
		res.is_p_ = false;

		if(lop.extp_->cat() == TCAT_INT) {
			if(rop.extp_->cat() == TCAT_INT) {
				mlc::icblock_append_delete(res.icb_p_, lop.icb_p_);
				mlc::icblock_append_delete(res.icb_p_, rop.icb_p_);
				res.icb_p_->append_instruction( new ai::MULI());

				res.extp_ = lop.extp_;

			} else if(rop.extp_->cat() == TCAT_REAL) {
				/* silent convert to real */
				mlc::icblock_append_delete(res.icb_p_, lop.icb_p_);
				res.icb_p_->append_instruction( new ai::CVRTIR);
				mlc::icblock_append_delete(res.icb_p_, rop.icb_p_);
				res.icb_p_->append_instruction( new ai::MULR());
				res.extp_ = rop.extp_;
				
			} else {
				error(DUERR_CANNOTCONVERT, l_op);
			}
		} else if(lop.extp_->cat() == TCAT_REAL) {
			if(rop.extp_->cat() == TCAT_INT) {
				mlc::icblock_append_delete(res.icb_p_, lop.icb_p_);
				/* silent convert to real */
				mlc::icblock_append_delete(res.icb_p_, rop.icb_p_);
				res.icb_p_->append_instruction( new ai::CVRTIR);
				res.icb_p_->append_instruction( new ai::MULR());
				res.extp_ = lop.extp_;

			} else if(rop.extp_->cat() == TCAT_REAL) {
				mlc::icblock_append_delete(res.icb_p_, lop.icb_p_);
				mlc::icblock_append_delete(res.icb_p_, rop.icb_p_);
				res.icb_p_->append_instruction( new ai::MULR());
				res.extp_ = rop.extp_;
				
			} else {
				error(DUERR_CANNOTCONVERT, l_op);
			}
		} else {
			error(DUERR_CANNOTCONVERT, l_op);

		}


	}

	void
	expr_sub( MlaskalCtx* ctx, MlaskalExpr & res, MlaskalExpr & lop, MlaskalExpr & rop, int l_op) {
		expr_deref(ctx, lop);
		expr_deref(ctx, rop);
		expr_deptr(lop);
		expr_deptr(rop);
		res.is_p_ = false;

		if(lop.extp_->cat() == TCAT_INT) {
			if(rop.extp_->cat() == TCAT_INT) {
				mlc::icblock_append_delete(res.icb_p_, lop.icb_p_);
				mlc::icblock_append_delete(res.icb_p_, rop.icb_p_);
				res.icb_p_->append_instruction( new ai::SUBI());

				res.extp_ = lop.extp_;

			} else if(rop.extp_->cat() == TCAT_REAL) {
				/* silent convert to real */
				mlc::icblock_append_delete(res.icb_p_, lop.icb_p_);
				res.icb_p_->append_instruction( new ai::CVRTIR);
				mlc::icblock_append_delete(res.icb_p_, rop.icb_p_);
				res.icb_p_->append_instruction( new ai::SUBR());
				res.extp_ = rop.extp_;
				
			} else {
				error(DUERR_CANNOTCONVERT, l_op);
			}
		} else if(lop.extp_->cat() == TCAT_REAL) {
			if(rop.extp_->cat() == TCAT_INT) {
				mlc::icblock_append_delete(res.icb_p_, lop.icb_p_);
				/* silent convert to real */
				mlc::icblock_append_delete(res.icb_p_, rop.icb_p_);
				res.icb_p_->append_instruction( new ai::CVRTIR);
				res.icb_p_->append_instruction( new ai::SUBR());
				res.extp_ = lop.extp_;

			} else if(rop.extp_->cat() == TCAT_REAL) {
				mlc::icblock_append_delete(res.icb_p_, lop.icb_p_);
				mlc::icblock_append_delete(res.icb_p_, rop.icb_p_);
				res.icb_p_->append_instruction( new ai::SUBR());
				res.extp_ = rop.extp_;
				
			} else {
				error(DUERR_CANNOTCONVERT, l_op);
			}
		} else {
			error(DUERR_CANNOTCONVERT, l_op);

		}


	}

	void
	expr_divr( MlaskalCtx* ctx, MlaskalExpr & res, MlaskalExpr & lop, MlaskalExpr & rop, int l_op) {
		expr_deref(ctx, lop);
		expr_deref(ctx, rop);
		expr_deptr(lop);
		expr_deptr(rop);
		res.is_p_ = false;

		if(lop.extp_->cat() == TCAT_INT) {
			if(rop.extp_->cat() == TCAT_INT) {
				mlc::icblock_append_delete(res.icb_p_, lop.icb_p_);
				res.icb_p_->append_instruction( new ai::CVRTIR);
				mlc::icblock_append_delete(res.icb_p_, rop.icb_p_);
				res.icb_p_->append_instruction( new ai::CVRTIR);
				res.icb_p_->append_instruction( new ai::DIVR());

				res.extp_ = ctx->tab->logical_real();

			} else if(rop.extp_->cat() == TCAT_REAL) {
				mlc::icblock_append_delete(res.icb_p_, lop.icb_p_);
				res.icb_p_->append_instruction( new ai::CVRTIR);
				mlc::icblock_append_delete(res.icb_p_, rop.icb_p_);
				res.icb_p_->append_instruction( new ai::DIVR());
				res.extp_ = rop.extp_;
				
			} else {
				error(DUERR_CANNOTCONVERT, l_op);
			}
		} else if(lop.extp_->cat() == TCAT_REAL) {
			if(rop.extp_->cat() == TCAT_INT) {
				mlc::icblock_append_delete(res.icb_p_, lop.icb_p_);
				/* silent convert to real */
				mlc::icblock_append_delete(res.icb_p_, rop.icb_p_);
				res.icb_p_->append_instruction( new ai::CVRTIR);
				res.icb_p_->append_instruction( new ai::DIVR());
				res.extp_ = lop.extp_;

			} else if(rop.extp_->cat() == TCAT_REAL) {
				mlc::icblock_append_delete(res.icb_p_, lop.icb_p_);
				mlc::icblock_append_delete(res.icb_p_, rop.icb_p_);
				res.icb_p_->append_instruction( new ai::DIVR());
				res.extp_ = rop.extp_;
				
			} else {
				error(DUERR_CANNOTCONVERT, l_op);
			}
		} else {
			error(DUERR_CANNOTCONVERT, l_op);

		}
	}

	void
	expr_divi( MlaskalCtx* ctx, MlaskalExpr & res, MlaskalExpr & lop, MlaskalExpr & rop, int l_op) {
		expr_deref(ctx, lop);
		expr_deref(ctx, rop);
		expr_deptr(lop);
		expr_deptr(rop);
		res.is_p_ = false;

		if(lop.extp_->cat() == TCAT_INT) {
			if(rop.extp_->cat() == TCAT_INT) {
				mlc::icblock_append_delete(res.icb_p_, lop.icb_p_);
				mlc::icblock_append_delete(res.icb_p_, rop.icb_p_);
				res.icb_p_->append_instruction( new ai::DIVI());

				res.extp_ = rop.extp_;
			} else {
				error(DUERR_CANNOTCONVERT, l_op);
			}
		} else {
			error(DUERR_CANNOTCONVERT, l_op);

		}
	}

	void
	expr_mod( MlaskalCtx* ctx, MlaskalExpr & res, MlaskalExpr & lop, MlaskalExpr & rop, int l_op) {
		expr_deref(ctx, lop);
		expr_deref(ctx, rop);
		expr_deptr(lop);
		expr_deptr(rop);
		res.is_p_ = false;

		if(lop.extp_->cat() == TCAT_INT) {
			if(rop.extp_->cat() == TCAT_INT) {
				mlc::icblock_append_delete(res.icb_p_, lop.icb_p_);
				mlc::icblock_append_delete(res.icb_p_, rop.icb_p_);
				res.icb_p_->append_instruction( new ai::MODI());

				res.extp_ = rop.extp_;
			} else {
				error(DUERR_CANNOTCONVERT, l_op);
			}
		} else {
			error(DUERR_CANNOTCONVERT, l_op);

		}
	}

	void
	expr_and( MlaskalCtx* ctx, MlaskalExpr & res, MlaskalExpr & lop, MlaskalExpr & rop, int l_op) {
		expr_deref(ctx, lop);
		expr_deref(ctx, rop);
		expr_deptr(lop);
		expr_deptr(rop);
		res.is_p_ = false;

		if((lop.extp_->cat() == TCAT_BOOL ) && (rop.extp_->cat() == TCAT_BOOL)) {
				mlc::icblock_append_delete(res.icb_p_, lop.icb_p_);
				mlc::icblock_append_delete(res.icb_p_, rop.icb_p_);
				res.icb_p_->append_instruction( new ai::AND());

				res.extp_ = ctx->tab->logical_bool();
		} else {
			error(DUERR_CANNOTCONVERT, l_op);

		}
	}

	void
	expr_or( MlaskalCtx* ctx, MlaskalExpr & res, MlaskalExpr & lop, MlaskalExpr & rop, int l_op) {
		expr_deref(ctx, lop);
		expr_deref(ctx, rop);
		expr_deptr(lop);
		expr_deptr(rop);
		res.is_p_ = false;

		if((lop.extp_->cat() == TCAT_BOOL ) && (rop.extp_->cat() == TCAT_BOOL)) {
				mlc::icblock_append_delete(res.icb_p_, lop.icb_p_);
				mlc::icblock_append_delete(res.icb_p_, rop.icb_p_);
				res.icb_p_->append_instruction( new ai::OR());

				res.extp_ = ctx->tab->logical_bool();
		} else {
			error(DUERR_CANNOTCONVERT, l_op);

		}
	}

	void
	add_xstx(MlaskalCtx* ctx, mlc::icblock_pointer icb_p, MlaskalExpr & arr, MlaskalExpr & expr) {
		expr_deref(ctx, arr);
		expr_deref(ctx, expr);


		if(arr.extp_->cat() == mlc::TCAT_INT) {
			/* loads value on stack */
			if(expr.extp_->cat() == TCAT_INT) {
				mlc::icblock_append_delete(icb_p, expr.icb_p_);
			} else if(expr.extp_->cat() == TCAT_REAL) {
				mlc::icblock_append_delete(icb_p, expr.icb_p_);
				icb_p->append_instruction( new ai::CVRTRI());
				error(DUERR_CONVERSION, arr.id_l_);
			} else {
				error(DUERR_CANNOTCONVERT, arr.id_l_);
			}

			/* loads pointer into array on stack */
			mlc::icblock_append_delete(icb_p, arr.icb_p_);

			icb_p->append_instruction( new ai::XSTI());
		} else if(arr.extp_->cat() == mlc::TCAT_REAL) {
			/* loads value on stack */
			if(expr.extp_->cat() == TCAT_INT) {
				mlc::icblock_append_delete(icb_p, expr.icb_p_);
				icb_p->append_instruction( new ai::CVRTIR());
			} else if(expr.extp_->cat() == TCAT_REAL) {
				mlc::icblock_append_delete(icb_p, expr.icb_p_);
			} else {
				error(DUERR_CANNOTCONVERT, arr.id_l_);
			}

			/* loads pointer into array on stack */
			mlc::icblock_append_delete(icb_p, arr.icb_p_);

			icb_p->append_instruction( new ai::XSTR());
		} else if((arr.extp_->cat() == mlc::TCAT_BOOL) && (expr.extp_->cat() == mlc::TCAT_BOOL)) {
			/* loads value on stack */
			mlc::icblock_append_delete(icb_p, expr.icb_p_);

			/* loads pointer into array on stack */
			mlc::icblock_append_delete(icb_p, arr.icb_p_);

			icb_p->append_instruction( new ai::XSTB());
		} else if((arr.extp_->cat() == mlc::TCAT_STR) && (expr.extp_->cat() == mlc::TCAT_STR)) {
			/* loads value on stack */
			mlc::icblock_append_delete(icb_p, expr.icb_p_);

			/* loads pointer into array on stack */
			mlc::icblock_append_delete(icb_p, arr.icb_p_);

			icb_p->append_instruction( new ai::XSTS());
		} else {
			if(mlc::identical_type(arr.extp_, expr.extp_)) {
				//on stack pointer to array
				//save as local variable
				mlc::icblock_append_delete(icb_p, expr.icb_p_);
				icb_p->append_instruction( new ai::SSTP(10));
				mlc::icblock_append_delete(icb_p, arr.icb_p_);
				icb_p->append_instruction( new ai::SSTP(20));

				for(int i = 0; i < expr.extp_->compute_size(); ++i) {
					icb_p->append_instruction( new ai::SLDP(11));
					if(expr.extp_->compute_final_ptype() == mlaskal::PTYPE_INTEGER) {
						icb_p->append_instruction( new ai::XLDI());
					} else {
						assert(false);
					}

					icb_p->append_instruction( new ai::SLDP(20));

					if(expr.extp_->compute_final_ptype() == mlaskal::PTYPE_INTEGER) {
						icb_p->append_instruction( new ai::XSTI());
					} else {
						assert(false);
					}

					//shift pointer
					icb_p->append_instruction( new ai::SLDP(11));
					icb_p->append_instruction( new ai::LDLITI(ctx->tab->one()));
					icb_p->append_instruction( new ai::ADDP());
					icb_p->append_instruction( new ai::SSTP(10));

					icb_p->append_instruction( new ai::SLDP(21));
					icb_p->append_instruction( new ai::LDLITI(ctx->tab->one()));
					icb_p->append_instruction( new ai::ADDP());
					icb_p->append_instruction( new ai::SSTP(20));
				}
				//expr.icb_p_->append_instruction( new ai::SLDP(11));
				//expr.icb_p_->append_instruction( new ai::SLDP(10));
			} else {
				error(DUERR_CANNOTCONVERT, arr.id_l_);
			}
		}
	}

	void
	add_xldx(MlaskalExpr & expr) {

		if(expr.extp_->cat() == TCAT_INT) {
			expr.icb_p_->append_instruction( new ai::XLDI());
		} else if(expr.extp_->cat() == TCAT_REAL) {
			expr.icb_p_->append_instruction( new ai::XLDR());
		} else if(expr.extp_->cat() == TCAT_BOOL) {
			expr.icb_p_->append_instruction( new ai::XLDB());
		} else if(expr.extp_->cat() == TCAT_STR) {
			expr.icb_p_->append_instruction( new ai::XLDS());
		} else {
			assert(false);
		}
	}

	void add_xstx_param(MlaskalCtx* ctx, mlc::icblock_pointer ib, mlc::parameter_by_reference_symbol_pointer param, MlaskalExpr& expr) {

		expr_deref(ctx, expr);
		mlc::icblock_append_delete(ib, expr.icb_p_);
		ib->append_instruction(new ai::LLDP(param->address()));

		if(param->access_typed()->type()->cat() == TCAT_INT) {
			ib->append_instruction(new ai::XSTI());
		} else if(param->access_typed()->type()->cat() == TCAT_REAL) {
			ib->append_instruction(new ai::XSTR());
		} else if(param->access_typed()->type()->cat() == TCAT_STR) {
			ib->append_instruction(new ai::XSTS());
		} else if(param->access_typed()->type()->cat() == TCAT_BOOL) {
			ib->append_instruction(new ai::XSTB());
		} else if(param->access_typed()->type()->cat() == TCAT_ARRAY) {
			if(mlc::identical_type(param->access_typed()->type(), expr.extp_)) {
				//on stack pointer to target array and pointer to source array above it
				//save as local variable


				ib->append_instruction( new ai::SSTP(19)); //target

				ib->append_instruction( new ai::SSTP(10)); //source

				for(int i = 0; i < expr.extp_->compute_size(); ++i) {
					ib->append_instruction( new ai::SLDP(11));
					if(expr.extp_->compute_final_ptype() == mlaskal::PTYPE_INTEGER) {
						ib->append_instruction( new ai::XLDI());
					} else {
						assert(false);
					}

					ib->append_instruction( new ai::SLDP(20));

					if(expr.extp_->compute_final_ptype() == mlaskal::PTYPE_INTEGER) {
						ib->append_instruction( new ai::XSTI());
					} else {
						assert(false);
					}

					//shift pointer
					ib->append_instruction( new ai::SLDP(11));
					ib->append_instruction( new ai::LDLITI(ctx->tab->one()));
					ib->append_instruction( new ai::ADDP());
					ib->append_instruction( new ai::SSTP(10));

					ib->append_instruction( new ai::SLDP(21));
					ib->append_instruction( new ai::LDLITI(ctx->tab->one()));
					ib->append_instruction( new ai::ADDP());
					ib->append_instruction( new ai::SSTP(20));
				}
			} else {
				error(DUERR_CANNOTCONVERT, expr.id_l_);
			}

		} else {
			assert(false);
		}
	}

	void
	add_minus(MlaskalExpr & expr, int l) {

		if(expr.extp_->cat() == TCAT_INT) {
			expr.icb_p_->append_instruction( new ai::MINUSI());
		} else if(expr.extp_->cat() == TCAT_REAL) {
			expr.icb_p_->append_instruction( new ai::MINUSR());
		} else {
			error( DUERR_CANNOTCONVERT, l);
		}

	}
	
	void
	arr_index(MlaskalCtx* ctx, MlaskalExpr & expr, expr_vec & indicies, bool appending) {

		if(expr.extp_->cat() != TCAT_ARRAY) {
			assert(false); 
			//TODO not indexing an array
		}

		mlc::array_type_pointer arr = expr.extp_->access_array();
	
		////

		bool first = true;
		for(expr_vec::reverse_iterator rit = indicies.rbegin(); rit != indicies.rend(); ++rit) {
			if(!first) {
				//recursion
				if(arr->element_type()->cat() == TCAT_ARRAY) {
				arr = arr->element_type()->access_array();
				} else {
					//TODO error, bad indexing
					assert(false);
				}
			}
								
			expr_deref(ctx, *rit);
			expr_deptr(*rit);
			if(rit->extp_->cat() != TCAT_INT) {
				error( DUERR_NOTORDINAL, expr.id_l_);
			}
			
			mlc::icblock_append_delete( expr.icb_p_, rit->icb_p_);
			if(arr->index_type()->cat() != ctx->tab->logical_integer()->cat()) {
				/* thus range */
				expr.icb_p_->append_instruction( new ai::LDLITI(arr->index_type()->access_range()->lowerBound()));
				expr.icb_p_->append_instruction( new ai::SUBI());
			}
				//correct start
			
			expr.icb_p_->append_instruction( new ai::LDLITI( ctx->tab->ls_int().add( arr->element_type()->compute_size())));
			expr.icb_p_->append_instruction( new ai::MULI());
			if(!first || appending) {
				expr.icb_p_->append_instruction( new ai::ADDI());
			}
			first = false;
		}
		expr.extp_ = arr->element_type();

	}


	void add_for( MlaskalCtx* ctx, mlc::icblock_pointer ib, mlc::ls_id_type::const_pointer id, int l_id, MlaskalExpr & init_expr, int direct , MlaskalExpr & term_expr, mlc::icblock_pointer body) {
		
		mlc::ic_label start = mlc::new_label( ctx); //for loop start
		mlc::ic_label end = mlc::new_label( ctx); //for loop end

		
		mlc::symbol_pointer sp = ctx->tab->find_symbol(id);

		//initialize loop counter
		expr_deref(ctx, init_expr);
		mlc::icblock_append_delete(ib, init_expr.icb_p_);
		if(sp->kind() == SKIND_LOCAL_VARIABLE) {
			if(sp->access_typed()->type()->cat() != TCAT_INT) {
				error(mlc::DUERR_FORNOTINTEGER, l_id, *id);
			}

			ib->append_instruction( new ai::LSTI(sp->access_local_variable()->address()));

		} else 	if(sp->kind() == SKIND_GLOBAL_VARIABLE) {
			if(sp->access_typed()->type()->cat() != TCAT_INT) {
				error(mlc::DUERR_FORNOTINTEGER, l_id, *id);
			}

			ib->append_instruction( new ai::GSTI(sp->access_global_variable()->address()));

		} else {
			error(mlc::DUERR_FORNOTLOCAL, l_id, *id);

		}

		ib->add_label(start);
		
		//add terminator
		expr_deref(ctx, term_expr);
		mlc::icblock_append_delete(ib, term_expr.icb_p_);
		//add counter
		if(sp->kind() == SKIND_LOCAL_VARIABLE) {
			ib->append_instruction( new ai::LLDI(sp->access_local_variable()->address()));
		} else {
			ib->append_instruction( new ai::GLDI(sp->access_global_variable()->address()));
		}

		//compare counter to terminator
		if(direct == DUTOKGE_DOWNTO) {
			ib->append_instruction( new ai::GTI());
		} else {
			ib->append_instruction( new ai::LTI());
		}
		ib->append_instruction_with_target( new ai::JT(ib->end()), end);

		mlc::icblock_append_delete(ib, body); // append loop body

		//make increment
		if(sp->kind() == SKIND_LOCAL_VARIABLE) {
			ib->append_instruction( new ai::LLDI(sp->access_local_variable()->address()));
			
			ib->append_instruction( new ai::LDLITI(ctx->tab->one()));
			if(direct == DUTOKGE_DOWNTO) {
				ib->append_instruction( new ai::SUBI());
			} else {
				ib->append_instruction( new ai::ADDI());
			}

			ib->append_instruction( new ai::LSTI(sp->access_local_variable()->address()));
		} else {
			ib->append_instruction( new ai::GLDI(sp->access_global_variable()->address()));
			
			ib->append_instruction( new ai::LDLITI(ctx->tab->one()));
			if(direct == DUTOKGE_DOWNTO) {
				ib->append_instruction( new ai::SUBI());
			} else {
				ib->append_instruction( new ai::ADDI());
			}

			ib->append_instruction( new ai::GSTI(sp->access_global_variable()->address()));
		}
		
		ib->append_instruction_with_target( new ai::JMP(ib->end()), start);

		ib->add_label( end);
	}

	void add_while( MlaskalCtx* ctx, mlc::icblock_pointer ib, MlaskalExpr & expr, int l_expr, mlc::icblock_pointer body) {

		expr_deref(ctx, expr);

		if(expr.extp_->cat() != TCAT_BOOL) {
			error(DUERR_CANNOTCONVERT, l_expr);
			return;
		}

		mlc::ic_label start = mlc::new_label( ctx); //for loop start
		mlc::ic_label end = mlc::new_label( ctx); //for loop end
	
		ib->add_label( start);
		mlc::icblock_append_delete(ib, expr.icb_p_);
		ib->append_instruction_with_target( new ai::JF(ib->end()), end);

		mlc::icblock_append_delete(ib, body);
		ib->append_instruction_with_target( new ai::JMP(ib->end()), start);
		ib->add_label( end);
	}

	void add_repeat( MlaskalCtx* ctx, mlc::icblock_pointer ib, mlc::icblock_pointer body, MlaskalExpr & expr, int l_expr) {

		mlc::ic_label start = mlc::new_label( ctx); //for loop start
	
		//append body
		ib->add_label( start);
		mlc::icblock_append_delete(ib, body);

		//append terminator T6000 ;-)
		expr_deref(ctx, expr);
		if(expr.extp_->cat() != TCAT_BOOL) {
			error(DUERR_CANNOTCONVERT, l_expr);
			return;
		}
		mlc::icblock_append_delete(ib, expr.icb_p_);
		ib->append_instruction_with_target( new ai::JF(ib->end()), start);
		
	}

	void add_if( MlaskalCtx* ctx, mlc::icblock_pointer ib, MlaskalExpr & expr, int l_expr, mlc::icblock_pointer stmt1, mlc::icblock_pointer stmt2) {
		mlc::ic_label else_stmt = mlc::new_label( ctx); //for loop start
		mlc::ic_label end = mlc::new_label( ctx); //for loop start

		expr_deref(ctx, expr);
		if(expr.extp_->cat() != TCAT_BOOL) {
			error(DUERR_CANNOTCONVERT, l_expr);
			return;
		}

		mlc::icblock_append_delete(ib, expr.icb_p_);
		ib->append_instruction_with_target( new ai::JF(ib->end()), else_stmt);

		mlc::icblock_append_delete(ib, stmt1);
		ib->append_instruction_with_target( new ai::JMP(ib->end()), end);
		ib->add_label(else_stmt);
		mlc::icblock_append_delete(ib, stmt2);
		ib->add_label(end);
	}

	void add_rel( MlaskalCtx* ctx, MlaskalExpr & lexpr, int oper, int l_oper, MlaskalExpr & rexpr) {

		expr_deref(ctx, rexpr);
		expr_deref(ctx, lexpr);

		if((lexpr.extp_->cat() == TCAT_INT) && (rexpr.extp_->cat() == TCAT_INT)) {
			mlc::icblock_append_delete(lexpr.icb_p_, rexpr.icb_p_);

			switch(oper) {
				case DUTOKGE_LT:
					lexpr.icb_p_->append_instruction( new ai::LTI);
					break;
				case DUTOKGE_LE:
					lexpr.icb_p_->append_instruction( new ai::LEI);
					break;
				case DUTOKGE_NE:
					lexpr.icb_p_->append_instruction( new ai::NEI);
					break;
				case DUTOKGE_GE:
					lexpr.icb_p_->append_instruction( new ai::GEI);
					break;
				case DUTOKGE_GT:
					lexpr.icb_p_->append_instruction( new ai::GTI);
					break;
			}

		} else if((lexpr.extp_->cat() == TCAT_INT) && (rexpr.extp_->cat() == TCAT_REAL)) {
			assert(false);
			mlc::icblock_append_delete(lexpr.icb_p_, rexpr.icb_p_);
		} else if((lexpr.extp_->cat() == TCAT_REAL) && (rexpr.extp_->cat() == TCAT_INT)) {
			assert(false);
			mlc::icblock_append_delete(lexpr.icb_p_, rexpr.icb_p_);
		} else if((lexpr.extp_->cat() == TCAT_REAL) && (rexpr.extp_->cat() == TCAT_REAL)) {
			assert(false);
			mlc::icblock_append_delete(lexpr.icb_p_, rexpr.icb_p_);
		} else if((lexpr.extp_->cat() == TCAT_STR) && (rexpr.extp_->cat() == TCAT_STR)) {
			assert(false);
			mlc::icblock_append_delete(lexpr.icb_p_, rexpr.icb_p_);
		} else if((lexpr.extp_->cat() == TCAT_BOOL) && (rexpr.extp_->cat() == TCAT_BOOL)) {
			assert(false);
			mlc::icblock_append_delete(lexpr.icb_p_, rexpr.icb_p_);
		} else {
			error(DUERR_CANNOTCONVERT, l_oper);
		}

		lexpr.extp_ = ctx->tab->logical_bool();

	}

	void add_eq( MlaskalCtx* ctx, MlaskalExpr & lexpr, int l_oper, MlaskalExpr & rexpr) {

		expr_deref(ctx, rexpr);
		expr_deref(ctx, lexpr);

		if((lexpr.extp_->cat() == TCAT_INT) && (rexpr.extp_->cat() == TCAT_INT)) {
			mlc::icblock_append_delete(lexpr.icb_p_, rexpr.icb_p_);
			lexpr.icb_p_->append_instruction( new ai::EQI());

		} else if((lexpr.extp_->cat() == TCAT_INT) && (rexpr.extp_->cat() == TCAT_REAL)) {
			lexpr.icb_p_->append_instruction( new ai::CVRTIR());
			mlc::icblock_append_delete(lexpr.icb_p_, rexpr.icb_p_);
			lexpr.icb_p_->append_instruction( new ai::EQR());

		} else if((lexpr.extp_->cat() == TCAT_REAL) && (rexpr.extp_->cat() == TCAT_INT)) {
			mlc::icblock_append_delete(lexpr.icb_p_, rexpr.icb_p_);
			lexpr.icb_p_->append_instruction( new ai::CVRTIR());
			lexpr.icb_p_->append_instruction( new ai::EQR());

		} else if((lexpr.extp_->cat() == TCAT_REAL) && (rexpr.extp_->cat() == TCAT_REAL)) {
			mlc::icblock_append_delete(lexpr.icb_p_, rexpr.icb_p_);
			lexpr.icb_p_->append_instruction( new ai::EQR());

		} else if((lexpr.extp_->cat() == TCAT_STR) && (rexpr.extp_->cat() == TCAT_STR)) {
			mlc::icblock_append_delete(lexpr.icb_p_, rexpr.icb_p_);
			lexpr.icb_p_->append_instruction( new ai::EQS());

		} else if((lexpr.extp_->cat() == TCAT_BOOL) && (rexpr.extp_->cat() == TCAT_BOOL)) {
			mlc::icblock_append_delete(lexpr.icb_p_, rexpr.icb_p_);
			lexpr.icb_p_->append_instruction( new ai::EQB());

		} else {
			error(DUERR_CANNOTCONVERT, l_oper);
		}

		lexpr.extp_ = ctx->tab->logical_bool();

	}

	/*void copy_array(MlaskalCtx* ctx, mlc::icblock_pointer, mlc::symbol_pointer target, int l_target, MlasklaExpr & source) {

		expr_deref(source);

		return;

		mlc::symbol_pointer rs = ctx->tab->find_symbol(source.id_ci_);
		
	}*/

}
